# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Module for parallelization using mpi4py.

Allows for easy parallelization in master/slaves mode with one master
submitting function or method calls to slaves.
Uses mpi4py if available, otherwise processes calls sequentially in one
process.

Examples:
=========

Save the following lines in ``demo_mpi.py`` and run::

    > mpirun -n 10 python demo_mpi.py

1. Use master/slaves parallelization with the Network class:

   .. literalinclude:: ../../../../examples/modules/mpi/network_large.py

2. Do a Monte Carlo simulation as master/slaves:

   .. literalinclude:: ../../../../examples/modules/mpi/network_mc.py

3. Do a parameter scan without communication with a master, and just save
   the results in files:

   .. literalinclude:: ../../../../examples/modules/mpi/network_scan_no_comm.py

"""

#
#  Imports
#

import sys
import time
import traceback

import numpy


# try to get the communicator object to see whether mpi is available:
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    """(mpi4py.MPI.Comm instance) MPI communicator."""
    available = True
    """(bool) indicates that slaves are available."""
except ImportError:
    available = False


class MPIException(Exception):
    def __init__(self, value):
        Exception.__init__()
        self.value = value

    def __str__(self):
        return repr(self.value)


# initialize:

if available:
    size = comm.size
    """(int) number of MPI nodes (master and slaves)."""
    rank = comm.rank
    """(int) rank of this MPI node (0 is the master)."""
    am_master = (rank == 0)
    """(bool) indicates that this MPI node is the master."""
    if size < 2:
        available = False
else:
    size = 1
    rank = 0
    am_master = True

am_slave = not am_master
"""(bool) indicates that this MPI node is a slave."""
n_slaves = size - 1
"""(int) no. of slaves available."""
start_time = time.time()
"""(float) starting time of this MPI node."""
stats = []
"""
(list of dictionaries)
stats[id] contains processing statistics for the last call with this id. Keys:

  - "id": id of the call
  - "rank": MPI node who processed the call
  - "this_time": wall time for processing the call
  - "time_over_est": quotient of actual over estimated wall time
  - "n_processed": no. of calls processed so far by that slave, including this
  - "total_time": total wall time until this call was finished
"""

# initialization on master with slaves:

if am_master:

    total_time_est = numpy.zeros(size)
    """
    (numpy array of ints)
    total_time_est[i] is the current estimate of the total time
    MPI slave i will work on already submitted calls.
    On slave i, only total_time_est[i] is available.
    """
    total_time_est[0] = numpy.inf
    queue = []
    """(list) ids of submitted calls"""
    assigned = {}
    """
    (dictionary)
    assigned[id] is the slave assigned to the call with that id.
    """
    slave_queue = [[] for i in range(0, size)]
    """
    (list of lists)
    slave_queue[i] contains the ids of calls assigned to slave i.
    """
    n_processed = numpy.zeros(size).astype("int")
    """
    (list of ints)
    n_processed[rank] is the total number of calls processed by MPI node rank.
    On slave i, only total_time[i] is available.
    """
    total_time = numpy.zeros(size)
    """
    (list of floats)
    total_time[rank] is the total wall time until that node finished its last
    call.  On slave i, only total_time[i] is available.
    """

    if not available:
        # dictionary for results:
        results = {}
        """
        (dictionary)
        if mpi is not available, the result of submit_call(..., id=a) will be
        cached in results[a] until get_result(a).
        """

    def submit_call(name_to_call, args=(), kwargs={},
                    module="__main__", time_est=1, id=None, slave=None):
        """
        Submit a call for parallel execution.

        If called by the master and slaves are available, the call is submitted
        to a slave for asynchronous execution.

        If called by a slave or if no slaves are available, the call is instead
        executed synchronously on this MPI node.

        **Examples:**

            1. Provide ids and time estimate explicitly:

               .. code-block:: python

                  for n in range(0,10):
                      mpi.submit_call("doit", (n,A[n]), id=n, time_est=n**2)

                  for n in range(0,10):
                      result[n] = mpi.get_result(n)

            2. Use generated ids stored in a list:

               .. code-block:: python

                  for n in range(0,10):
                      ids.append(mpi.submit_call("doit", (n,A[n])))

                  for n in range(0,10):
                      results.append(mpi.get_result(ids.pop()))

            3. Ignore ids altogether:

               .. code-block:: python

                  for n in range(0,10):
                      mpi.submit_call("doit", (n,A[n]))

                  for n in range(0,10):
                      results.append(mpi.get_next_result())

            4. Call a module function and use keyword arguments:

               .. code-block:: python

                  mpi.submit_call("solve", (), {"a":a, "b":b},
                      module="numpy.linalg")

            5. Call a static class method from a package:

               .. code-block:: python

                  mpi.submit_call("Network._get_histogram", (values, n_bins),
                      module="pyunicorn")

              Note that it is module="pyunicorn" and not
              module="pyunicorn.network" here.

        :arg str name_to_call: name of callable object (usually a function or
            static method of a class) as contained in the namespace specified
            by module.
        :arg tuple args: the positional arguments to provide to the callable
            object.  Tuples of length 1 must be written (arg,).  Default: ()
        :arg dict kwargs: the keyword arguments to provide to the callable
            object.  Default: {}
        :arg str module: optional name of the imported module or submodule in
            whose namespace the callable object is contained. For objects
            defined on the script level, this is "__main__", for objects
            defined in an imported package, this is the package name. Must be a
            key of the dictionary sys.modules (check there after import if in
            doubt).  Default: "__main__"
        :arg float time_est: estimated relative completion time for this call;
            used to find a suitable slave. Default: 1
        :type id: object or None
        :arg  id: unique id for this call. Must be a possible dictionary key.
            If None, a random id is assigned and returned. Can be re-used after
            get_result() for this is. Default: None
        :type slave: int > 0 and < mpi.size, or None
        :arg  slave: optional no. of slave to assign the call to. If None, the
            call is assigned to the slave with the smallest current total time
            estimate. Default: None
        :return object: id of call, to be used in get_result().
        """
        if id is None:
            id = numpy.random.uniform()
        if id in assigned:
            raise MPIException("id ", str(id), " already in queue!")
        if slave is not None and am_slave:
            raise MPIException(
                "only the master can use slave= in submit_call()")
        if slave is None or slave < 1 or slave >= size:
            # find slave with least estimated total time:
            slave = numpy.argmin(total_time_est)
        if available:
            # send name to call, args, time_est to slave:
            if _verbose:
                print(f"MPI master : assigning call with id {id} to slave "
                      f"{slave}: {name_to_call} {args} {kwargs} ...")
            comm.send((name_to_call, args, kwargs, module, time_est),
                      dest=slave)
        else:
            # do it myself right now:
            slave = 0
            if _verbose:
                print(f"MPI master : calling {name_to_call} {args} {kwargs} "
                      "...")
            try:
                object_to_call = eval(name_to_call,
                                      sys.modules[module].__dict__)
            except NameError:
                sys.stderr.write(str(sys.modules[module].__dict__.keys()))
                raise
            call_time = time.time()
            results[id] = object_to_call(*args, **kwargs)
            this_time = time.time() - call_time
            n_processed[0] += 1
            total_time[0] = time.time() - start_time
            stats.append({"id": id, "rank": 0,
                          "this_time": this_time,
                          "time_over_est": this_time / time_est,
                          "n_processed": n_processed[0],
                          "total_time": total_time[0]})

        total_time_est[slave] += time_est
        queue.append(id)
        slave_queue[slave].append(id)
        assigned[id] = slave
        return id

    def get_result(id):
        """
        Return result of earlier submitted call.

        Can only be called by the master.

        If the call is not yet finished, waits for it to finish.
        Results should be collected in the same order as calls were submitted.
        For each slave, the results of calls assigned to that slave must be
        collected in the same order as those calls were submitted.
        Can only be called once per call.

        :type id: object
        :arg  id: id of an earlier submitted call, as provided to or returned
                  by submit_call().

        :rtype:  object
        :return: return value of call.
        """
        source = assigned[id]
        if available:
            if slave_queue[source][0] != id:
                raise MPIException("get_result(" + str(id)
                                   + ") called before get_result("
                                   + str(slave_queue[source][0]) + ")!")
            if _verbose:
                print(f"MPI master : retrieving result for call with id {id} "
                      f"from slave {source} ...")
            (result, this_stats) = comm.recv(source=source)
            stats.append(this_stats)
            n_processed[source] = this_stats["n_processed"]
            total_time[source] = this_stats["total_time"]
        else:
            if _verbose:
                print(f"MPI master : returning result for call with id {id} "
                      "...")
            result = results[id]
            # TODO: rather return a copy and del the original?
        queue.remove(id)
        slave_queue[source].remove(id)
        assigned.pop(id)
        return result

    def get_next_result():
        """
        Return result of next earlier submitted call whose result has not yet
        been got.

        Can only be called by the master.

        If the call is not yet finished, waits for it to finish.

        :rtype:  object
        :return: return value of call, or None of there are no more calls in
                 the queue.
        """
        if len(queue) > 0:
            id = queue[0]
            return get_result(id)
        else:
            return None

    def info():
        """
        Print processing statistics.

        Can only be called by the master.
        """

        call_times = numpy.array([s["this_time"] for s in stats])
        call_quotients = numpy.array([s["time_over_est"] for s in stats])

        if available:
            slave_quotients = total_time/total_time_est
            print("\n"
                  "MPI: processing statistics\n"
                  "     =====================\n"
                  "     results collected:         "
                  f"{n_processed[1:].sum()}\n"
                  "     results not yet collected: "
                  f"{len(queue)}\n"
                  "     total reported time:       "
                  f"{call_times.sum()}\n"
                  "     mean time per call:        "
                  f"{call_times.mean()}\n"
                  "     std.dev. of time per call: "
                  f"{call_times.std()}\n"
                  "     coeff. of var. of actual over estd. time per call: "
                  f"{call_quotients.std()/call_quotients.mean()}\n"
                  "     slaves:                      "
                  f"{n_slaves}\n"
                  "     mean calls per slave:        "
                  f"{n_processed[1:].mean()}\n"
                  "     std.dev. of calls per slave: "
                  f"{n_processed[1:].std()}\n"
                  "     min calls per slave:         "
                  f"{n_processed[1:].min()}\n"
                  "     max calls per slave:         "
                  f"{n_processed[1:].max()}\n"
                  "     mean time per slave:        "
                  f"{total_time.mean()}\n"
                  "     std.dev. of time per slave: "
                  f"{total_time.std()}\n"
                  "     coeff. of var. of actual over estd. time per slave: "
                  f"{slave_quotients.std()/slave_quotients.mean()}\n")
        else:
            print("\n"
                  "MPI: processing statistics\n"
                  "     =====================\n"
                  "     results collected:         "
                  f"{n_processed[0]}\n"
                  "     results not yet collected: "
                  f"{len(queue)}\n"
                  "     total reported time:       "
                  f"{call_times.sum()}\n"
                  "     mean time per call:        "
                  f"{call_times.mean()}\n"
                  "     std.dev. of time per call: "
                  f"{call_times.std()}\n"
                  "     coeff. of var. of actual over estd. time per call: "
                  f"{call_quotients.std()/call_quotients.mean()}\n")

    def terminate():
        """
        Tell all slaves to terminate.

        Can only be called by the master.
        """
        global available
        if available:
            # tell slaves to terminate:
            for slave in range(1, size):
                if _verbose:
                    print(f"MPI master : telling slave {slave} "
                          "to terminate...")
                comm.send(("terminate", (), {}, "", 0), dest=slave)
            available = False

    def abort():
        """
        Abort execution on all MPI nodes immediately.

        Can be called by master and slaves.
        """
        traceback.print_exc()
        if _verbose:
            print("MPI master : aborting...")
        comm.Abort()

else:  # am_slave and available:

    total_time_est = numpy.zeros(size)*numpy.nan
    total_time_est[rank] = 0
    n_processed = numpy.zeros(size)*numpy.nan
    n_processed[rank] = 0
    total_time = numpy.zeros(size)*numpy.nan
    total_time[rank] = 0

    def serve():
        """
        Serve submitted calls until told to finish.

        Can only be called by slaves.

        Call this function from inside your definition of slave() if slaves
        need to perform initializations different from the master, like this:

        >>> def slave():
        >>>     do = whatever + initialization - is * necessary
        >>>     mpi.serve()
        >>>     do = whatever + cleanup - is * necessary

        If you don't define slave(), serve() will be called automatically by
        mpi.run().
        """
        if _verbose:
            print("MPI slave ", rank, ": waiting for calls.")
        # wait for orders:
        while True:
            # get next task from queue:
            (name_to_call, args, kwargs, module, time_est) = \
                comm.recv(source=0)
            # TODO: add some timeout and check whether master lives!
            if name_to_call == "terminate":
                if _verbose:
                    print("MPI slave", rank, ": terminating...")
                break
            if _verbose:
                print(f"MPI slave {rank}: calling {name_to_call} {args} ...")
            try:
                object_to_call = eval(name_to_call,
                                      sys.modules[module].__dict__)
            except NameError:
                sys.stderr.write(str(sys.modules[module].__dict__.keys()))
                raise
            total_time_est[rank] += time_est
            call_time = time.time()
            result = object_to_call(*args, **kwargs)
            this_time = time.time() - call_time
            n_processed[rank] += 1
            stats.append({"id": id, "rank": rank,
                          "this_time": this_time,
                          "time_over_est": this_time / time_est,
                          "n_processed": n_processed[rank],
                          "total_time": time.time() - start_time})
            if _verbose:
                print("MPI slave", rank, ": sending result...")
            comm.send((result, stats[-1]), dest=0)

    def abort():
        traceback.print_exc()
        if _verbose:
            print("MPI slave", rank, ": aborting...")
        comm.Abort()

    # TODO: update total_time_est at return time

_verbose = False


def run(verbose=False):
    """
    Run in master/slaves mode until master() finishes.

    Must be called on all MPI nodes after function master() was defined.

    On the master, run() calls master() and returns when master() returns.

    On each slave, run() calls slave() if that is defined, or calls serve()
    otherwise, and returns when slave() returns, or when master() returns on
    the master, or when master calls terminate().

    :arg bool verbose: whether processing information should be printed.
    """

    # transfer verbose into global environment:
    global _verbose
    _verbose = verbose
    """
    (bool) indicated whether processing information should be printed.
    """

    _globals = sys.modules['__main__'].__dict__

    if available:  # run in mpi mode
        if am_master:  # I'm master
            if verbose:
                print("MPI master : started, using", size-1, "slaves.")
            try:  # put everything in a try block to be able to terminate!
                if "master" in _globals:
                    _globals["master"]()
                else:
                    print("MPI master : function master() not found!")
            except ValueError:
                abort()
            terminate()
            if verbose:
                print("MPI master : finished.")
        else:  # I'm slave
            if "slave" in _globals:
                _globals["slave"]()
            else:
                serve()
    else:  # run as single processor
        if verbose:
            print("MPI master : not available, running as a single process.")
        if "master" in _globals:
            _globals["master"]()
        else:
            print("MPI master : function master() not found!")
        if verbose:
            print("MPI master : finished.")



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the JointRecurrencePlot class.
"""

import numpy as np

import pytest

from pyunicorn.timeseries import JointRecurrencePlot
from pyunicorn.funcnet import CouplingAnalysis


@pytest.mark.parametrize("n", [2, 10, 50])
def test_recurrence(metric: str, n: int):
    ts = CouplingAnalysis.test_data()[:n, 0]
    jrp = JointRecurrencePlot(ts, ts, threshold=(.1, .1),
                              metric=(metric, metric))
    dist = {}
    for i in "xy":
        jrp.embedding = getattr(jrp, f"{i}_embedded")
        dist[i] = jrp.distance_matrix(metric=metric)
    assert all(d.shape == (n, n) for d in dist.values())
    assert np.allclose(dist["x"], dist["y"])
    assert jrp.recurrence_matrix().shape == (n, n)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

cimport cython

import random

import numpy as np
cimport numpy as cnp
from numpy cimport ndarray, abs
import numpy.random as rd
randint = rd.randint

from ...core._ext.types import NODE, DEGREE, FIELD, DFIELD
from ...core._ext.types cimport \
    ADJ_t, MASK_t, NODE_t, DEGREE_t, WEIGHT_t, DWEIGHT_t, FIELD_t, DFIELD_t

cdef extern from "src_numerics.c":
    void _do_nsi_hamming_clustering_fast(int n2, int nActiveIndices,
        float mind0, float minwp0, int lastunited, int part1, int part2,
        double *distances, int *theActiveIndices, double *linkedWeights,
        double *weightProducts, double *errors, double *result, int *mayJoin)
    double _vertex_current_flow_betweenness_fast(int N, double Is, double It,
        float *admittance, float *R, int i)
    void _edge_current_flow_betweenness_fast(int N, double Is, double It,
        float *admittance, float *R, float *ECFB)


# geo_network =================================================================


# parameters for `_randomly_rewire_geomodel()`
ctypedef bint (*rewire_cond_len)(FIELD_t[:,:], float, int, int, int, int)
ctypedef bint (*rewire_cond_deg)(DEGREE_t[:], int, int, int, int)

cdef:
    # condition C1
    inline bint cond_len_c1(
        FIELD_t[:,:] D, float eps, int s, int t, int k, int l):
        return (
            (abs(D[s,t] - D[k,t]) < eps and abs(D[k,l] - D[s,l]) < eps) or
            (abs(D[s,t] - D[s,l]) < eps and abs(D[k,l] - D[k,t]) < eps))
    # condition C2
    inline bint cond_len_c2(
        FIELD_t[:,:] D, float eps, int s, int t, int k, int l):
        return (
            abs(D[s,t] - D[s,l]) < eps and abs(D[t,s] - D[t,k]) < eps and
            abs(D[k,l] - D[k,t]) < eps and abs(D[l,k] - D[l,s]) < eps)
    # invariance of degree-degree correlations
    inline bint cond_deg_corr(DEGREE_t[:] degree, int s, int t, int k, int l):
        return (degree[s] == degree[k] and degree[t] == degree[l])
    # tautology
    rewire_cond_deg cond_deg_true = NULL


cdef void _randomly_rewire_geomodel(int iterations, float eps,
    ndarray[ADJ_t, ndim=2] A, ndarray[FIELD_t, ndim=2] D, int E,
    ndarray[NODE_t, ndim=2] edges, ndarray[DEGREE_t, ndim=1] degree,
    rewire_cond_len cond_len, rewire_cond_deg cond_deg):

    cdef int i = 0, s, t, k, l, edge1, edge2

    while (i < iterations):
        # Randomly choose 2 edges
        edge1 = np.floor(rd.random() * E)
        edge2 = np.floor(rd.random() * E)
        s, t = edges[edge1,[0,1]]
        k, l = edges[edge2,[0,1]]

        # Proceed only if old links are disjoint
        if ((s != k and s != l and t != k and t != l) and
            # Proceed only if new links do NOT already exist
            (A[s,l] == 0 and A[t,k] == 0) and
            # Proceed only if link conditions are fulfilled
            (cond_deg is NULL or cond_deg(degree, s, t, k, l)) and
            cond_len(D, eps, s, t, k, l)):

                # Now rewire the links symmetrically & increment i
                A[s,t] = A[t,s] = 0
                A[k,l] = A[l,k] = 0
                A[s,l] = A[l,s] = 1
                A[t,k] = A[k,t] = 1
                edges[edge1,[0,1]] = s, l
                edges[edge2,[0,1]] = k, t

                i+=1


def _randomly_rewire_geomodel_I(int iterations, float eps,
    ndarray[ADJ_t, ndim=2] A, ndarray[FIELD_t, ndim=2] D, int E,
    ndarray[NODE_t, ndim=2] edges):
    cdef:
        ndarray[DEGREE_t, ndim=1] null = np.array([], dtype=DEGREE)
    _randomly_rewire_geomodel(iterations, eps, A, D, E, edges, null,
                              cond_len_c1, cond_deg_true)

def _randomly_rewire_geomodel_II(int iterations, float eps,
    ndarray[ADJ_t, ndim=2] A, ndarray[FIELD_t, ndim=2] D, int E,
    ndarray[NODE_t, ndim=2] edges):
    cdef:
        ndarray[DEGREE_t, ndim=1] null = np.array([], dtype=DEGREE)
    _randomly_rewire_geomodel(iterations, eps, A, D, E, edges, null,
                              cond_len_c2, cond_deg_true)

def _randomly_rewire_geomodel_III(int iterations, float eps,
    ndarray[ADJ_t, ndim=2] A, ndarray[FIELD_t, ndim=2] D, int E,
    ndarray[NODE_t, ndim=2] edges, ndarray[DEGREE_t, ndim=1] degree):
    _randomly_rewire_geomodel(iterations, eps, A, D, E, edges, degree,
                              cond_len_c2, cond_deg_corr)


# interacting_networks ========================================================


cdef void overwriteAdjacency(
    ndarray[ADJ_t, ndim=2] A, ndarray[ADJ_t, ndim=2] cross_A,
    ndarray[NODE_t, ndim=1] nodes1, ndarray[NODE_t, ndim=1] nodes2,
    int m, int n):
    """
    Overwrite the adjacency matrix of the full interacting network with the
    randomly rewired cross edges of the two considered subnetworks.
    """
    cdef:
        int i, j
        NODE_t n1, n2

    for i in range(m):
        for j in range(n):
            n1, n2 = nodes1[i], nodes2[j]
            A[n1, n2] = A[n2, n1] = cross_A[i, j]


def _randomlySetCrossLinks(
    ndarray[ADJ_t, ndim=2] A, ndarray[ADJ_t, ndim=2] cross_A,
    int number_cross_links,
    ndarray[NODE_t, ndim=1] nodes1, ndarray[NODE_t, ndim=1] nodes2,
    int m, int n):
    """
    >>> A = np.eye(2, dtype=np.int)
    >>> _randomlySetCrossLinks(A, np.array([[0]]), 1,
    ...                        np.array([0]), np.array([1]), 1, 1)
    >>> np.all(A == np.ones(2))
    True
    """
    cdef:
        int i, j

    # create random cross links
    for _ in range(number_cross_links):
        while True:
            i, j = randint(m), randint(n)
            if not cross_A[i, j]:
                break
        cross_A[i, j] = 1
    overwriteAdjacency(A, cross_A, nodes1, nodes2, m, n)


def _randomlyRewireCrossLinks(
    ndarray[ADJ_t, ndim=2] A,
    ndarray[ADJ_t, ndim=2] cross_A,
    ndarray[NODE_t, ndim=2] cross_links,
    ndarray[NODE_t, ndim=1] nodes1,
    ndarray[NODE_t, ndim=1] nodes2,
    int number_cross_links, int number_swaps):

    cdef:
        int m = int(len(nodes1)), n = int(len(nodes2))
        NODE_t e1, e2, a, b, c, d

    # implement permutations
    for _ in range(number_swaps):
        while True:
            # choose two random edges
            e1, e2 = randint(number_cross_links), randint(number_cross_links)
            a, b = cross_links[e1]
            c, d = cross_links[e2]
            # repeat the procedure in case there already exists a link between
            # starting point of e1 and ending point of e2 or vice versa
            if not (cross_A[a, d] or cross_A[c, b]):
                break
        # delete initial edges within the cross adjacency matrix
        cross_A[a, b] = cross_A[c, d] = 0
        # create new edges within the cross adjacency matrix by swapping the
        # ending points of e1 and e2
        cross_A[a, d] = cross_A[c, b] = 1
        # likewise, adjust cross_links
        b                  = cross_links[e1, 1]
        cross_links[e1, 1] = cross_links[e2, 1]
        cross_links[e2, 1] = b
    overwriteAdjacency(A, cross_A, nodes1, nodes2, m, n)


def _cross_transitivity(
    ndarray[ADJ_t, ndim=2] A,
    ndarray[NODE_t, ndim=1] nodes1, ndarray[NODE_t, ndim=1] nodes2):

    cdef:
        int m = int(len(nodes1)), n = int(len(nodes2))
        int i, j, k
        NODE_t n1, n2, n3
        long triangles = 0, triples = 0

    for i in range(m):
        n1 = nodes1[i]
        # loop over unique pairs of nodes in subnetwork 2
        for j in range(n):
            n2 = nodes2[j]
            if A[n1, n2]:
                for k in range(j):
                    n3 = nodes2[k]
                    if A[n1, n3]:
                        triples += 1
                    if A[n2, n3] and A[n3, n1]:
                        triangles += 1
    if triples:
        return triangles / float(triples)
    else:
        return 0.0


def _nsi_cross_transitivity(
    ndarray[ADJ_t, ndim=2] A,
    ndarray[NODE_t, ndim=1] nodes1,
    ndarray[NODE_t, ndim=1] nodes2,
    ndarray[DWEIGHT_t, ndim=1] node_weights):

    cdef:
        int m = int(len(nodes1)), n = int(len(nodes2))
        int v, p, q
        NODE_t node_v, node_p, node_q
        DWEIGHT_t weight_v, weight_p, ppv, pqv, T1 = 0, T2 = 0

    for v in range(m):
        node_v = nodes1[v]
        weight_v = node_weights[node_v]
        for p in range(n):
            node_p = nodes2[p]
            if A[node_v, node_p]:
                weight_p = node_weights[node_p]
                ppv = weight_p * weight_p * weight_v
                T1 += ppv
                T2 += ppv
                for q in range(p + 1, n):
                    node_q = nodes2[q]
                    if A[node_v, node_q]:
                        pqv = 2 * weight_p * node_weights[node_q] * weight_v
                        T2 += pqv
                        if A[node_p, node_q]:
                            T1 += pqv
    return T1 / T2


def _cross_local_clustering(
    ndarray[ADJ_t, ndim=2] A,
    ndarray[DFIELD_t, ndim=1] norm,
    ndarray[NODE_t, ndim=1] nodes1, ndarray[NODE_t, ndim=1] nodes2,
    ndarray[DFIELD_t, ndim=1] cross_clustering):

    cdef:
        int m = int(len(nodes1)), n = int(len(nodes2))
        int i, j, k
        NODE_t n1, n2, n3
        long counter

    for i in range(m):
        n1 = nodes1[i]
        # check if node1[i] has cross degree larger than 1
        if norm[i]:
            counter = 0
            # loop over unique pairs of nodes in subnetwork 2
            for j in range(n):
                n2 = nodes2[j]
                if A[n1, n2]:
                    for k in range(j):
                        n3 = nodes2[k]
                        if A[n2, n3] and A[n3, n1]:
                            counter += 1
            cross_clustering[i] = counter / norm[i]


def _nsi_cross_local_clustering(
    ndarray[ADJ_t, ndim=2] A,
    ndarray[DFIELD_t, ndim=1] nsi_cc,
    ndarray[NODE_t, ndim=1] nodes1, ndarray[NODE_t, ndim=1] nodes2,
    ndarray[DWEIGHT_t, ndim=1] node_weights):

    cdef:
        int m = int(len(nodes1)), n = int(len(nodes2))
        int v, p, q
        NODE_t node_v, node_p, node_q
        DWEIGHT_t weight_p

    for v in range(m):
        node_v = nodes1[v]
        for p in range(n):
            node_p = nodes2[p]
            if A[node_v, node_p]:
                weight_p = node_weights[node_p]
                nsi_cc[v] += weight_p * weight_p
                for q in range(p + 1, n):
                    node_q = nodes2[q]
                    if A[node_p, node_q] and A[node_q, node_v]:
                        nsi_cc[v] += <WEIGHT_t>2 * weight_p * node_weights[node_q]


# network =====================================================================


def _local_cliquishness_4thorder(
    int N, ndarray[ADJ_t, ndim=2] A, ndarray[DEGREE_t, ndim=1] degree):

    cdef:
        int i, j, k, l
        int index
        NODE_t node1, node2, node3, degree_i, order = 4
        long counter
        ndarray[NODE_t, ndim=1] neighbors = np.zeros(N, dtype=NODE)
        ndarray[DFIELD_t, ndim=1] local_cliquishness = \
            np.zeros(N, dtype=DFIELD)

    # Iterate over all nodes
    for i in range(N):
        # If degree is smaller than order - 1, set local cliquishness to 0
        degree_i = degree[i]
        if degree_i >= order - 1:
            # Get neighbors of node i
            index = 0
            for j in range(N):
                if A[i, j] == 1:
                    neighbors[index] = j
                    index += 1
            counter = 0
            # Iterate over possibly existing edges between 3 neighbors of i
            for j in range(degree_i):
                node1 = neighbors[j]
                for k in range(degree_i):
                    node2 = neighbors[k]
                    if A[node1, node2] == 1:
                        for l in range(degree_i):
                            node3 = neighbors[l]
                            if A[node2, node3] == 1 and A[node3, node1] == 1:
                                counter += 1
            local_cliquishness[i] = counter /\
                (degree_i * (degree_i - 1) * (degree_i - 2))
    return local_cliquishness


def _local_cliquishness_5thorder(
    int N, ndarray[ADJ_t, ndim=2] A, ndarray[DEGREE_t, ndim=1] degree):

    cdef:
        int i, index
        NODE_t j, k, l, m, node1, node2, node3, node4, degree_i, order = 5
        long counter
        ndarray[NODE_t, ndim=1] neighbors = np.zeros(N, dtype=NODE)
        ndarray[DFIELD_t, ndim=1] local_cliquishness = \
            np.zeros(N, dtype=DFIELD)

    # Iterate over all nodes
    for i in range(N):
        # If degree is smaller than order - 1, set local cliquishness to 0
        degree_i = degree[i]
        if degree_i >= order - 1:
            # Get neighbors of node i
            index = 0
            for j in range(N):
                if A[i, j] == 1:
                    neighbors[index] = j
                    index += 1
            counter = 0
            # Iterate over possibly existing edges between 4 neighbors of i
            for j in range(degree_i):
                node1 = neighbors[j]
                for k in range(degree_i):
                    node2 = neighbors[k]
                    if A[node1, node2] == 1:
                        for l in range(degree_i):
                            node3 = neighbors[l]
                            if A[node1, node3] == 1 and A[node2, node3] == 1:
                                for m in range(degree_i):
                                    node4 = neighbors[m]
                                    if (A[node1, node4] == 1 and
                                        A[node2, node4] == 1 and
                                        A[node3, node4] == 1):
                                        counter += 1
            local_cliquishness[i] = counter /\
                (degree_i * (degree_i - 1) * (degree_i - 2) * (degree_i -3))
    return local_cliquishness


def _nsi_betweenness(
    int N, ndarray[DWEIGHT_t, ndim=1] w,
    ndarray[DEGREE_t, ndim=1] k,
    ndarray[NODE_t, ndim=1] flat_neighbors,
    ndarray[MASK_t, ndim=1] is_source,
    ndarray[NODE_t, ndim=1] targets):
    """
    Performs Newman's algorithm. [Newman2001]_
    """

    cdef:
        long int E = len(flat_neighbors)
        int j, qi, oi, queue_len, l_index, ql
        NODE_t l, i, next_d, dl, ol, fi
        DFIELD_t base_factor
        ndarray[NODE_t, ndim=1] offsets = np.zeros(N, dtype=NODE)
        ndarray[NODE_t, ndim=1] distances_to_j = np.ones(N, dtype=NODE)
        ndarray[NODE_t, ndim=1] n_predecessors = np.zeros(N, dtype=NODE)
        ndarray[NODE_t, ndim=1] flat_predecessors = np.zeros(E, dtype=NODE)
        ndarray[NODE_t, ndim=1] queue = np.zeros(N, dtype=NODE)
        ndarray[DFIELD_t, ndim=1] multiplicity_to_j = np.zeros(N, dtype=DFIELD)
        ndarray[DFIELD_t, ndim=1] betweenness_to_j = np.zeros(N, dtype=DFIELD)
        ndarray[DFIELD_t, ndim=1] excess_to_j = np.zeros(N, dtype=DFIELD)
        ndarray[DFIELD_t, ndim=1] betweenness_times_w = np.zeros(N, dtype=DFIELD)

    # init node offsets
    # NOTE: We don't use k.cumsum() since that uses too much memory!
    for i in range(1, N):
        offsets[i] = offsets[i-1] + k[i-1]

    for j in targets:
        # init distances to j and queue of nodes by distance from j
        distances_to_j.fill(2 * N)
        n_predecessors.fill(0)
        flat_predecessors.fill(0)
        queue.fill(0)
        multiplicity_to_j.fill(0)

        # init contribution of paths ending in j to the betweenness of l
        for l in range(N):
            excess_to_j[l] = betweenness_to_j[l] = is_source[l] * w[l]

        distances_to_j[j] = 0
        queue[0] = j
        queue_len = 1
        multiplicity_to_j[j] = w[j]

        # process the queue forward and grow it on the way: (this is the
        # standard breadth-first search giving all the shortest paths to j)
        qi = 0
        while qi < queue_len:
            i = queue[qi]
            if i == -1:
                # this should never happen ...
                print("Opps: %d,%d,%d\n" % qi, queue_len, i)
                break
            next_d = distances_to_j[i] + 1
            # iterate through all neighbors l of i
            oi = offsets[i]
            for l_index in range(oi, oi+k[i]):
                # if on a shortest j-l-path, register i as predecessor of l
                l = flat_neighbors[l_index]
                dl = distances_to_j[l]
                if dl >= next_d:
                    fi = offsets[l] + n_predecessors[l]
                    n_predecessors[l] += 1
                    flat_predecessors[fi] = i
                    multiplicity_to_j[l] += w[l] * multiplicity_to_j[i]
                    if dl > next_d:
                        distances_to_j[l] = next_d
                        queue[queue_len] = l
                        queue_len += 1
            qi += 1

        # process the queue again backward: (this is Newman's 2nd part where the
        # contribution of paths ending in j to the betweenness of all nodes is
        # computed recursively by traversing the shortest paths backwards)
        for ql in range(queue_len-1, -1, -1):
            l = queue[ql]
            if l == -1:
                print("Opps: %d,%d,%d\n" % ql, queue_len, l)
                break
            if l == j:
                # set betweenness and excess to zero
                betweenness_to_j[l] = excess_to_j[l] = 0
            else:
                # otherwise, iterate through all predecessors i of l:
                base_factor = w[l] / multiplicity_to_j[l]
                ol = offsets[l]
                for fi in range(ol, ol+n_predecessors[l]):
                    # add betweenness to predecessor
                    i = flat_predecessors[fi]
                    betweenness_to_j[i] += betweenness_to_j[l] * base_factor * \
                        multiplicity_to_j[i]

        betweenness_times_w += w[j] * (betweenness_to_j - excess_to_j)
    return betweenness_times_w


def _mpi_newman_betweenness(
    ndarray[ADJ_t, ndim=2] this_A, ndarray[DFIELD_t, ndim=2] V,
    int N, int start_i, int end_i):
    """
    This function does the outer loop for a certain range start_i-end_i of
    c's.  it gets the full V matrix but only the needed rows of the A matrix.
    Each parallel job will consist of a call to this function:
    """

    cdef:
        int i_rel, j, s, t, i_abs
        double sum_s, sum_j, Vis_minus_Vjs

        int this_N = end_i - start_i
        ndarray[DFIELD_t, ndim=1] this_betweenness = \
            np.zeros(this_N, dtype=DFIELD)

    for i_rel in range(this_N):
        # correct i index for V matrix
        i_abs = i_rel + start_i
        for j in range(N):
             if this_A[i_rel, j]:
                sum_j = 0.0
                for s in range(N):
                    if i_abs != s:
                        Vis_minus_Vjs = V[i_abs, s] - V[j, s]
                        sum_s = 0.0
                        for t in range(s):
                            if i_abs != t:
                                sum_s += abs(Vis_minus_Vjs - V[i_abs, t] +
                                             V[j, t])
                        sum_j += sum_s
                this_betweenness[i_rel] += sum_j

    return this_betweenness, start_i, end_i


def _mpi_nsi_newman_betweenness(
    ndarray[ADJ_t, ndim=2] this_A,
    ndarray[DFIELD_t, ndim=2] V, int N, ndarray[DWEIGHT_t, ndim=1] w,
    ndarray[MASK_t, ndim=2] this_not_adj_or_equal, int start_i, int end_i):

    cdef:
        int i_rel, j, s, t, i_abs
        double sum_s, sum_j, Vis_minus_Vjs

        int this_N = end_i - start_i
        ndarray[DFIELD_t, ndim=1] this_betweenness =\
            np.zeros(this_N, dtype=DFIELD)

    for i_rel in range(this_N):
        i_abs = i_rel + start_i
        for j in range(N):
             if this_A[i_rel, j]:
                sum_j = 0.0
                for s in range(N):
                    if this_not_adj_or_equal[i_rel, s]:
                        Vis_minus_Vjs = V[i_abs, s] - V[j, s]
                        sum_s = 0.0
                        for t in range(s):
                            if this_not_adj_or_equal[i_rel, t]:
                                sum_s += w[t] *\
                                    abs(Vis_minus_Vjs - V[i_abs, t] + V[j, t])
                        sum_j += w[s] * sum_s
                this_betweenness[i_rel] += w[j] * sum_j

    return this_betweenness, start_i, end_i


def _do_nsi_clustering_I(
    int n_cands, ndarray[NODE_t, ndim=1] cands,
    ndarray[DEGREE_t, ndim=1] D_cluster,
    ndarray[DFIELD_t, ndim=1] w, double d0,
    ndarray[NODE_t, ndim=1] D_firstpos, ndarray[NODE_t, ndim=1] D_nextpos,
    int N, dict dict_D, dict dict_Delta):

    cdef:
        int ca, ij, i, j, posh, h, ih, jh
        double wi, wj, wc, wjd0, Delta_inc, wh, whd0, Dih, Djh, Dch_wc, \
            Dch_wc_Dih_wi, Dch_wc_Djh_wj, Dch_wc_whd0

    # loop thru candidates:
    for ca in range(n_cands):
        ij = cands[ca]
        i = int(ij/N)
        j = ij%N
        wi = w[i]
        wj = w[j]
        wc = wi + wj
        wjd0 = wj * d0
        Delta_inc = 0.0
        # loop thru all nbs of i other than j:
        posh = D_firstpos[i]
        while (posh > 0):
            h = D_cluster[posh]
            if (h != j):
                ih = N*i+h
                jh = N*j+h
                wh = w[h]
                Dih = dict_D[ih]
                if (dict_D.has_key(jh)):
                    Djh = dict_D[jh]
                else:
                    Djh = wh * wjd0

                Dch_wc = (Dih + Djh) / wc
                Dch_wc_Dih_wi = Dch_wc - Dih/wi
                Dch_wc_Djh_wj = Dch_wc - Djh/wj

                Delta_inc += (wi * Dch_wc_Dih_wi*Dch_wc_Dih_wi +
                              wj * Dch_wc_Djh_wj*Dch_wc_Djh_wj) / wh

            posh = D_nextpos[posh]

        # loop thru all nbs of j other than i which are not nbs of i:
        posh = D_firstpos[j]
        while (posh > 0):
            h = D_cluster[posh]
            ih = N*i+h
            wh = w[h]
            whd0 = wh * d0
            if (h != i and not dict_D.has_key(ih)):
                jh = N*j+h
                Djh = dict_D[jh]
                Dch_wc = (wi*whd0 + Djh) / wc
                Dch_wc_whd0 = Dch_wc - whd0
                Dch_wc_Djh_wj = Dch_wc - Djh/wj

                Delta_inc += (wi * Dch_wc_whd0*Dch_wc_whd0 +
                              wj * Dch_wc_Djh_wj*Dch_wc_Djh_wj) / wh

            posh = D_nextpos[posh]

        dict_Delta[ij] = float(dict_Delta[ij]) + Delta_inc

    return dict_Delta


def _do_nsi_clustering_II(int a, int b,
    ndarray[DEGREE_t, ndim=1] D_cluster,
    ndarray[DFIELD_t, ndim=1] w, double d0,
    ndarray[NODE_t, ndim=1] D_firstpos, ndarray[NODE_t, ndim=1] D_nextpos,
    int N, dict dict_D, dict dict_Delta):

    cdef:
        double  wa = w[a], wb = w[b], wc = wa+wb, wad0 = wa*d0, wbd0 = wb*d0, \
                wa1, wa1sq, wa1d0, Da1a1, Da1a, Da1b, Da1c, wb1, wb1d0, wb1sq, \
                wa1b1, wc1, Db1b1, Da1b1, Dc1c1_wc1sq, \
                Dc1c1_wc1sq_Da1a1_wa1sq, Dc1c1_wc1sq_Da1b1_wa1b1, \
                Dc1c1_wc1sq_Db1b1_wb1sq, \
                Delta_new, Delta_inc, wc2 = 0, Da1c2, Db1c2, \
                Dc1c2_wc1, Dc1c2_wc1_Da1c2_wa1, Dc1c2_wc1_Db1c2_wb1, \
                Dc1c2_wc1_wc2_d0, Db1a, Db1b, Db1c, Dc1a_wc1, Dc1b_wc1, \
                Dc1c_wc1, Dc1c_wc1_Da1c_wa1, Dc1a_wc1_Da1a_wa1, \
                Dc1b_wc1_Da1b_wa1, Dc1c_wc1_Db1c_wb1, Dc1a_wc1_Db1a_wb1, \
                Dc1b_wc1_Db1b_wb1
        int N1 = N+1, posa1 = D_firstpos[a] # a meaning c!
        int a1, a1N, a1a1, a1a, a1b, posb1, b1, b1N, posc2, b1c2 = 0, b1a, b1b, \
            a1b1key, b1b1
        DEGREE_t c2

    while (posa1 > 0):
        a1 = D_cluster[posa1]
        a1N = a1*N
        a1a1 = a1*N1
        a1a = a1N+a
        a1b = a1N+b
        wa1 = w[a1]
        wa1sq = wa1*wa1
        wa1d0 = wa1*d0
        if (dict_D.has_key(a1a1)):
            Da1a1 = dict_D[a1a1]
        else:
            Da1a1 = 0.0
        if (dict_D.has_key(a1a)):
            Da1a = dict_D[a1a]
        else:
            Da1a = wa1*wad0
        if (dict_D.has_key(a1b)):
            Da1b = dict_D[a1b]
        else:
            Da1b = wa1*wbd0

        Da1c = Da1a + Da1b
        posb1 = D_firstpos[a1]

        while (posb1 > 0):
            b1 = D_cluster[posb1]
            b1N = b1*N
            b1b1 = b1*N1

            if (a1 < b1):
                a1b1key = a1N+b1
            else:
                a1b1key = b1N+a1
            if (dict_Delta.has_key(a1b1key)):
                wb1 = w[b1]
                wb1d0 = wb1 * d0
                wb1sq = wb1*wb1
                wa1b1 = wa1*wb1
                wc1 = wa1+wb1
                if (dict_D.has_key(b1b1)):
                    Db1b1 = dict_D[b1b1]
                else:
                    Db1b1 = 0.0
                if (dict_D.has_key(a1b1key)):
                    Da1b1 = dict_D[a1b1key]
                else:
                    Da1b1 = wb1 * wa1d0
                Dc1c1_wc1sq = (Da1a1+Db1b1+<float>2*Da1b1) / (wc1*wc1)
                if (b1 == a): # a meaning c!
                    Dc1c1_wc1sq_Da1a1_wa1sq = Dc1c1_wc1sq - Da1a1/wa1sq
                    Dc1c1_wc1sq_Db1b1_wb1sq = Dc1c1_wc1sq - Db1b1/wb1sq
                    Dc1c1_wc1sq_Da1b1_wa1b1 = Dc1c1_wc1sq - Da1b1/wa1b1
                    Delta_new = wa1sq * Dc1c1_wc1sq_Da1a1_wa1sq * \
                                Dc1c1_wc1sq_Da1a1_wa1sq + \
                                wb1sq * Dc1c1_wc1sq_Db1b1_wb1sq * \
                                Dc1c1_wc1sq_Db1b1_wb1sq + \
                                <float>2 * wa1b1 * Dc1c1_wc1sq_Da1b1_wa1b1 * \
                                Dc1c1_wc1sq_Da1b1_wa1b1
                    # loop thru all nbs c2 of a1 other than b1:
                    posc2 = D_firstpos[a1]
                    while (posc2 > 0):
                        c2 = D_cluster[posc2]
                        if (c2 != b1):
                            b1c2 = b1N+c2
                            wc2 = w[c2]
                            Da1c2 = dict_D[a1N+c2]
                            if (dict_D.has_key(b1c2)):
                                Db1c2 = dict_D[b1c2]
                            else:
                                Db1c2 = wc2*wb1d0
                            Dc1c2_wc1 = (Da1c2 + Db1c2) / wc1
                            Dc1c2_wc1_Da1c2_wa1 = Dc1c2_wc1 - Da1c2/wa1
                            Dc1c2_wc1_Db1c2_wb1 = Dc1c2_wc1 - Db1c2/wb1
                            Delta_new += (wa1 * Dc1c2_wc1_Da1c2_wa1 * \
                                          Dc1c2_wc1_Da1c2_wa1 + \
                                          wb1 * Dc1c2_wc1_Db1c2_wb1 * \
                                          Dc1c2_wc1_Db1c2_wb1) / wc2

                        posc2 = D_nextpos[posc2]

                    #  loop thru all nbs of b1 other than a1
                    #  which are not nbs of a1:
                    posc2 = D_firstpos[b1]
                    while (posc2 > 0):
                        c2 = D_cluster[posc2]
                        if (c2 != a1):
                            if not (dict_D.has_key(a1N+c2)):
                                b1c2 = b1N+c2
                                wc2 = w[c2]

                            if (dict_D.has_key(b1c2)):
                                Db1c2 = dict_D[b1c2]
                            else:
                                Db1c2 = wc2*wb1d0
                            Dc1c2_wc1 = (wc2*wa1d0 + Db1c2) / wc1
                            Dc1c2_wc1_wc2_d0 = Dc1c2_wc1 - wc2*d0
                            Dc1c2_wc1_Db1c2_wb1 = Dc1c2_wc1 - Db1c2/wb1
                            Delta_new += (wa1 * Dc1c2_wc1_wc2_d0 * \
                                          Dc1c2_wc1_wc2_d0 + \
                                          wb1 * Dc1c2_wc1_Db1c2_wb1 * \
                                          Dc1c2_wc1_Db1c2_wb1) / wc2

                        posc2 = D_nextpos[posc2]
                    dict_Delta[a1b1key] = Delta_new

                else:
                    b1a = b1N+a
                    b1b = b1N+b
                    if (dict_D.has_key(b1a)):
                        Db1a = dict_D[b1a]
                    else:
                        Db1a = wb1*wad0
                    if (dict_D.has_key(b1b)):
                        Db1b = dict_D[b1b]
                    else:
                        Db1b = wb1*wbd0;
                    Db1c = Db1a + Db1b
                    wc1 = wa1 + wb1
                    Dc1a_wc1 = (Da1a + Db1a) / wc1
                    Dc1b_wc1 = (Da1b + Db1b) / wc1
                    Dc1c_wc1 = (Da1c + Db1c) / wc1
                    Dc1c_wc1_Da1c_wa1 = Dc1c_wc1 - Da1c/wa1
                    Dc1a_wc1_Da1a_wa1 = Dc1a_wc1 - Da1a/wa1
                    Dc1b_wc1_Da1b_wa1 = Dc1b_wc1 - Da1b/wa1
                    Dc1c_wc1_Db1c_wb1 = Dc1c_wc1 - Db1c/wb1
                    Dc1a_wc1_Db1a_wb1 = Dc1a_wc1 - Db1a/wb1
                    Dc1b_wc1_Db1b_wb1 = Dc1b_wc1 - Db1b/wb1
                    Delta_inc = (Dc1c_wc1_Da1c_wa1*Dc1c_wc1_Da1c_wa1/wc - \
                                 Dc1a_wc1_Da1a_wa1*Dc1a_wc1_Da1a_wa1/wa - \
                                 Dc1b_wc1_Da1b_wa1*Dc1b_wc1_Da1b_wa1/wb)/ \
                                wa1 + \
                                (Dc1c_wc1_Db1c_wb1*Dc1c_wc1_Db1c_wb1/wc - \
                                 Dc1a_wc1_Db1a_wb1*Dc1a_wc1_Db1a_wb1/wa - \
                                 Dc1b_wc1_Db1b_wb1*Dc1b_wc1_Db1b_wb1/wb)/ \
                                wb1
                    Delta_inc *= <float>2
                    dict_Delta[a1b1key] = float(dict_Delta[a1b1key])+Delta_inc

            posb1 = D_nextpos[posb1]

        posa1 = D_nextpos[posa1]

    return dict_Delta


def _do_nsi_hamming_clustering(int n2, int nActiveIndices, float mind0,
    float minwp0, int lastunited, int part1, int part2,
    ndarray[DFIELD_t, ndim=2] distances,
    ndarray[NODE_t, ndim=1] theActiveIndices,
    ndarray[DFIELD_t, ndim=2] linkedWeights,
    ndarray[DFIELD_t, ndim=2] weightProducts,
    ndarray[DFIELD_t, ndim=2] errors,
    ndarray[DFIELD_t, ndim=1] results,
    ndarray[MASK_t, ndim=2] mayJoin):

    _do_nsi_hamming_clustering_fast(n2, nActiveIndices, mind0, minwp0,
        lastunited, part1, part2,
        <DFIELD_t*> cnp.PyArray_DATA(distances),
        <int*> cnp.PyArray_DATA(theActiveIndices),
        <DFIELD_t*> cnp.PyArray_DATA(linkedWeights),
        <DFIELD_t*> cnp.PyArray_DATA(weightProducts),
        <DFIELD_t*> cnp.PyArray_DATA(errors),
        <DFIELD_t*> cnp.PyArray_DATA(results),
        <int*> cnp.PyArray_DATA(mayJoin))


# grid ========================================================================


def _calculate_angular_distance(
    ndarray[FIELD_t, ndim=1] cos_lat,
    ndarray[FIELD_t, ndim=1] sin_lat,
    ndarray[FIELD_t, ndim=1] cos_lon,
    ndarray[FIELD_t, ndim=1] sin_lon,
    ndarray[FIELD_t, ndim=2] cosangdist, int N):

    cdef:
        FIELD_t expr
        int i,j

    for i in range(N):
        for j in range(i+1):
            expr = sin_lat[i]*sin_lat[j] + cos_lat[i]*cos_lat[j] * \
                (sin_lon[i]*sin_lon[j] + cos_lon[i]*cos_lon[j])

            if expr > 1:
                expr = 1
            elif expr < -1:
                expr = -1

            cosangdist[i, j] = cosangdist[j, i] = expr


def _calculate_euclidean_distance(
    ndarray[FIELD_t, ndim=2] x,
    ndarray[FIELD_t, ndim=2] distance,
    int N_dim, int N_nodes):

    cdef:
        int i,j,k
        FIELD_t expr

    for i in range(N_nodes):
        for j in range(i+1):
            expr = 0
            for k in range(N_dim):
                expr += (x[k, i]-x[k, j])**2
            distance[i, j] = distance[j, i] = expr**(<FIELD_t> 0.5)


# resistive_network ===========================================================


def _vertex_current_flow_betweenness(int N, double Is, double It,
    ndarray[FIELD_t, ndim=2] admittance, ndarray[FIELD_t, ndim=2] R,
    int i):

    return _vertex_current_flow_betweenness_fast(N, Is, It,
        <FIELD_t*> cnp.PyArray_DATA(admittance),
        <FIELD_t*> cnp.PyArray_DATA(R), i)


def _edge_current_flow_betweenness(int N, double Is, double It,
    ndarray[FIELD_t, ndim=2] admittance,
    ndarray[FIELD_t, ndim=2] R,):

    # alloc output
    cdef ndarray[FIELD_t, ndim=2, mode='c'] ECFB = \
            np.zeros((N, N), dtype=FIELD)

    _edge_current_flow_betweenness_fast(N, Is, It,
        <FIELD_t*> cnp.PyArray_DATA(admittance),
        <FIELD_t*> cnp.PyArray_DATA(R),
        <FIELD_t*> cnp.PyArray_DATA(ECFB))

    return ECFB



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for analyzing spatially embedded complex networks, handling
multivariate data and generating time series surrogates.

Written by Jakob Runge.
CMSI Method Reference: [Pompe2011]_
"""

# array object and fast numerics
import numpy


#
#  Define class CouplingAnalysisPurePython
#

class CouplingAnalysisPurePython:

    """
    Contains methods to calculate coupling matrices from large arrays
    of scalar time series.

    Comprises linear and information theoretic measures, lagged
    and directed (causal) couplings.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, dataarray, only_tri=False, silence_level=0):
        """
        Initialize an instance of CouplingAnalysisPurePython.

        Possible choices for only_tri:
          - "True" will calculate only the upper triangle of the coupling
            matrix, excluding the diagonal, assuming symmetry (not for directed
            measures)
          - "False" will calculate the whole matrix (asymmetry somes from
             different integration ranges)

        :type dataarray: 4D, 3D or 2D Numpy array [time, index, index] or
                         [time, index]
        :arg dataarray: The time series array with time in first dimension
        :arg bool only_tri: Symmetric/asymmetric assumption on coupling matrix.
        :arg int silence_level: The inverse level of verbosity of the object.
        """

        #  only_tri will calculate the upper triangle excluding the diagonal
        #  only. This assumes stationarity on the time series
        self.only_tri = only_tri

        #  Set silence level
        self.silence_level = silence_level

        #  Flatten observable anomaly array along lon/lat dimension to allow
        #  for more convinient indexing and transpose the whole array as this
        #  is faster in loops
        if numpy.ndim(dataarray) == 4:
            (self.total_time, n_lev, n_lat, n_lon) = dataarray.shape
            self.N = n_lev * n_lat * n_lon
            self.dataarray = dataarray.reshape(-1, self.N).T.copy()
        if numpy.ndim(dataarray) == 3:
            (self.total_time, n_lat, n_lon) = dataarray.shape
            self.N = n_lat * n_lon
            self.dataarray = dataarray.reshape(-1, self.N).T.copy()

        elif numpy.ndim(dataarray) == 2:
            (self.total_time, self.N) = dataarray.shape
            self.dataarray = dataarray.T.copy()

        else:
            print("irregular array shape...")
            self.dataarray = dataarray.T.copy()

        #  factorials below 10 in a list for permutation patterns
        self.factorial = \
            numpy.array([1, 1, 2, 6, 24, 120, 720, 5040, 40320, 362880])
        self.patternized = False
        self.has_fft = False
        self.originalFFT = None

        #  lag_mode dict
        self.lag_modi = {"all": 0, "sum": 1, "max": 2}

    def __str__(self):
        """
        Return a string representation of the CouplingAnalysisPurePython
        object.
        """
        shape = self.dataarray.shape
        return (f'CouplingAnalysisPurePython: {shape[0]} variables, '
                f'{shape[1]} timesteps.')

    #
    #  Define methods to calculate correlation strength and lags
    #

    #
    #  Routines for calculating Cross Correlation
    #

    def cross_correlation(self, tau_max=0, lag_mode='all'):
        """
        Returns the normalized cross correlation from all pairs of nodes from
        a range of time lags.

        The calculation ranges are shown below::

            (-------------------------total_time--------------------------)
            (---tau_max---)(---------corr_range------------)(---tau_max---)

        CC is calculated about corr_range and with the other time series
        shifted by tau

        Possible choices for lag_mode:

        - "all" will return the full function for all lags, possible large
          memory need if only_tri is True, only the upper triangle contains the
          values, the lower one is zeros
        - "sum" will return the sum over positive and negative lags seperatly,
          each inclunding tau=0 corrmat[0] is the positive sum, corrmat[1] the
          negative sum
        - "max" will return only the maximum coupling (in corrmat[0]) and its
          lag (in corrmat[1])

        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: the output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """
        #  Normalize anomaly time series to zero mean and unit variance for all
        #  lags, array contains normalizations for all lags
        corr_range = self.total_time - 2*tau_max
        normalized_array = numpy.empty((2*tau_max + 1, self.N, corr_range),
                                       dtype="float32")

        for t in range(2*tau_max + 1):
            #  Remove mean value from time series at each vertex (grid point)
            normalized_array[t] = self.dataarray[:, t:t+corr_range] - \
                self.dataarray[:, t:t+corr_range].\
                mean(axis=1).reshape(self.N, 1)

            #  Normalize the variance of anomalies to one
            normalized_array[t] /= normalized_array[t].\
                std(axis=1).reshape(self.N, 1)

            #  Correct for grid points with zero variance in their time series
            normalized_array[t][numpy.isnan(normalized_array[t])] = 0

        return self._calculate_cc(normalized_array, corr_range=corr_range,
                                  tau_max=tau_max, lag_mode=lag_mode)

    def shuffled_surrogate_for_cc(self, fourier=False, tau_max=1,
                                  lag_mode='all'):
        """
        Returns a correlation matrix calculated with an independently shuffled
        surrogate of the dataarray of length corr_range for all taus.

        :arg int corr_range: length of sample
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """
        corr_range = self.total_time - 2*tau_max

        # Shuffle a copy of dataarray separatly for each node
        array = numpy.copy(self.dataarray)
        if fourier:
            array = self.correlatedNoiseSurrogates(array)
        else:
            for i in range(self.N):
                numpy.random.shuffle(array[i])

        sample_array = numpy.zeros((1, self.N, corr_range), dtype="float32")

        sample_array[0] = array[:, :corr_range]
        sample_array[0] -= sample_array[0].mean(axis=1).reshape(self.N, 1)
        sample_array[0] /= sample_array[0].std(axis=1).reshape(self.N, 1)
        sample_array[0, numpy.isnan(sample_array[0])] = 0

        res = self._calculate_cc(sample_array, corr_range=corr_range,
                                 tau_max=0, lag_mode='all')

        if lag_mode == 'all':
            corrmat = numpy.repeat(res, 2*tau_max + 1, axis=0)
        elif lag_mode == 'sum':
            corrmat = numpy.array([abs(res[0]), abs(res[0])]) * (tau_max+1.)
        elif lag_mode == 'max':
            corrmat = numpy.array([abs(res[0]),
                                   numpy.random.randint(-tau_max, tau_max+1,
                                                        (self.N, self.N))])

        return corrmat

    def time_surrogate_for_cc(self, sample_range=100, tau_max=1,
                              lag_mode='all'):
        """
        Returns a joint shuffled surrogate of the full dataarray of length
        sample_range for all taus.

        Used for time evolution analysis. First one initializes the
        CouplingAnalysis class with the full dataarray and then this function
        is called for every single surrogate.

        :arg int sample_range: length of sample
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """

        perm = numpy.random.permutation(
            range(tau_max, self.total_time - tau_max))[:sample_range]

        sample_array = numpy.empty((2*tau_max + 1, self.N, sample_range),
                                   dtype="float32")

        for t in range(2 * tau_max + 1):
            tau = t - tau_max
            sample_array[t] = self.dataarray[:, perm + tau]
            sample_array[t] -= sample_array[t].mean(axis=1).reshape(self.N, 1)
            sample_array[t] /= sample_array[t].std(axis=1).reshape(self.N, 1)
            sample_array[t][numpy.isnan(sample_array[t])] = 0

        return self._calculate_cc(sample_array, corr_range=sample_range,
                                  tau_max=tau_max, lag_mode=lag_mode)

    def _calculate_cc(self, array, corr_range, tau_max, lag_mode):
        """
        Returns the CC matrix.

        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices

        ## lag_mode dict
        mode = self.lag_modi[lag_mode]
        """

        # lag_mode dict
        mode = self.lag_modi[lag_mode]
        only_tri = int(self.only_tri)

        if lag_mode == 'all':
            corrmat = numpy.zeros((2*tau_max + 1, self.N, self.N),
                                  dtype='float32')
        elif lag_mode == 'sum':
            corrmat = numpy.zeros((2, self.N, self.N), dtype='float32')
        elif lag_mode == 'max':
            corrmat = numpy.zeros((2, self.N, self.N), dtype='float32')

        # loop over all node pairs, NOT symmetric due to time shifts!
        for i in range(self.N-only_tri):
            for j in range((i+1)*only_tri, self.N):

                if mode == 2:
                    maxcross = 0.0
                    argmax = 0

                # loop over taus INCLUDING the last tau value
                for t in range(2*tau_max+1):

                    # here the actual cross correlation is calculated
                    crossij = (array[tau_max, i, :] * array[t, j, :]).mean()

                    # fill in values in matrix depending on lag_mode
                    if mode == 0:
                        corrmat[t, i, j] = crossij

                    elif mode == 1:
                        if t <= tau_max:
                            corrmat[1, i, j] += numpy.abs(crossij)
                        if t >= tau_max:
                            corrmat[0, i, j] += numpy.abs(crossij)

                    elif mode == 2:
                        # calculate max and argmax by comparing to previous
                        # value and storing max
                        if numpy.abs(crossij) > maxcross:
                            maxcross = numpy.abs(crossij)
                            argmax = t

                if mode == 2:
                    corrmat[0, i, j] = maxcross
                    corrmat[1, i, j] = argmax - tau_max

        if self.only_tri:
            if lag_mode == 'all':
                corrmat = corrmat + corrmat.transpose(0, 2, 1)[::-1]
            elif lag_mode == 'sum':
                corrmat[0] += corrmat[1].transpose()
                corrmat[1] = corrmat[0].transpose()
            elif lag_mode == 'max':
                corrmat[0] += corrmat[0].transpose()
                corrmat[1] -= corrmat[1].transpose()

        return corrmat

    #
    #  Routines for calculating Mutual Information with adaptive bins
    #

    def mutual_information(self, bins=16, tau_max=0, lag_mode='all'):
        """
        Returns the normalized mutual information from all pairs of nodes from
        a range of time lags.

        MI = H_x + H_y - H_xy

        Uses adaptive bins, where each marginal bin contains the same number of
        samples. Then the marginal entropies have equal probable distributions
        H_x = H_y = log(bins)

        The calculation ranges are shown below::

            (-------------------------total_time--------------------------)
            (---tau_max---)(---------corr_range------------)(---tau_max---)

        MI is calculated about corr_range and with the other time series
        shifted by tau

        Possible choices for lag_mode:

        - "all" will return the full function for all lags, possible large
          memory need if only_tri is True, only the upper triangle contains the
          values, the lower one is zeros
        - "sum" will return the sum over positive and negative lags seperatly,
          each inclunding tau=0 corrmat[0] is the positive sum, corrmat[1] the
          negative sum
        - "max" will return only the maximum coupling (in corrmat[0]) and its
          lag (in corrmat[1])

        :arg int bins: number of bins for estimating MI
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """
        if bins < 255:
            dtype = 'uint8'
        else:
            dtype = 'int16'

        # Normalize anomaly time series to zero mean and unit variance for all
        # lags, array contains normalizations for all lags
        corr_range = self.total_time - 2*tau_max

        # get the bin quantile steps
        bin_edge = numpy.ceil(corr_range/float(bins))

        symbolic_array = numpy.empty((2*tau_max + 1, self.N, corr_range),
                                     dtype=dtype)

        for t in range(2*tau_max + 1):

            array = self.dataarray[:, t:t+corr_range]

            # get the lower edges of the bins for every time series
            edges = numpy.sort(array, axis=1)[:, ::bin_edge]
            bins = edges.shape[1]

            # This gives the symbolic time series
            symbolic_array[t] = \
                (array.reshape(self.N, corr_range, 1)
                 >= edges.reshape(self.N, 1, bins)).sum(axis=2) - 1

        return self._calculate_mi(symbolic_array, corr_range=corr_range,
                                  bins=bins, tau_max=tau_max,
                                  lag_mode=lag_mode)

    def mutual_information_edges(self, bins=16, tau=0, lag_mode='all'):
        """
        Returns the normalized mutual information from all pairs of nodes from
        a range of time lags.

        MI = H_x + H_y - H_xy

        Uses adaptive bins, where each marginal bin contains the same number of
        samples. Then the marginal entropies have equal probable distributions
        H_x = H_y = log(bins)

        The calculation ranges are shown below::

            (-------------------------total_time--------------------------)
            (---tau_max---)(---------corr_range------------)(---tau_max---)

        MI is calculated about corr_range and with the other time series
        shifted by tau

        Possible choices for lag_mode:

        - "all" will return the full function for all lags, possible large
          memory need if only_tri is True, only the upper triangle contains the
          values, the lower one is zeros
        - "sum" will return the sum over positive and negative lags seperatly,
          each inclunding tau=0 corrmat[0] is the positive sum, corrmat[1] the
          negative sum
        - "max" will return only the maximum coupling (in corrmat[0]) and its
          lag (in corrmat[1])

        :arg int bins: number of bins for estimating MI
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 2D numpy array (float) [index, index]
        :return: bin edges for zero lag
        """

        # get the bin quantile steps
        bin_edge = numpy.ceil(self.total_time/float(bins))

        array = self.dataarray[:, :]
        array[:-tau, 1] = array[tau, 1]

        # get the lower edges of the bins for every time series
        edges = numpy.sort(array, axis=1)[:, ::bin_edge]
        bins = edges.shape[1]

        return edges

    def shuffled_surrogate_for_mi(self, fourier=False, bins=16, tau_max=0,
                                  lag_mode='all'):
        """
        Returns a shuffled surrogate of normalized mutual information from all
        pairs of nodes from a range of time lags.

        :arg int bins: number of bins for estimating MI
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """
        if bins < 255:
            dtype = 'uint8'
        else:
            dtype = 'int16'

        #  Normalize anomaly time series to zero mean and unit variance for all
        #  lags, array contains normalizations for all lags
        corr_range = self.total_time - 2*tau_max

        # Shuffle a copy of dataarray seperatly for each node
        array = numpy.copy(self.dataarray)
        if fourier:
            array = self.correlatedNoiseSurrogates(array)
        else:
            for i in range(self.N):
                numpy.random.shuffle(array[i])

        # get the bin quantile steps
        bin_edge = numpy.ceil(corr_range/float(bins))

        symbolic_array = numpy.empty((1, self.N, corr_range), dtype=dtype)

        array = array[:, :corr_range]

        # get the lower edges of the bins for every time series
        edges = numpy.sort(array, axis=1)[:, ::bin_edge]
        bins = edges.shape[1]

        # This gives the symbolic time series
        symbolic_array[0] = \
            (array.reshape(self.N, corr_range, 1)
             >= edges.reshape(self.N, 1, bins)).sum(axis=2) - 1

        res = self._calculate_mi(symbolic_array, corr_range=corr_range,
                                 bins=bins, tau_max=0, lag_mode='all')

        if lag_mode == 'all':
            corrmat = numpy.repeat(res, 2*tau_max + 1, axis=0)
        elif lag_mode == 'sum':
            corrmat = numpy.array([res[0], res[0]]) * (tau_max+1.)
        elif lag_mode == 'max':
            corrmat = numpy.array(
                [res[0], numpy.random.randint(-tau_max, tau_max+1,
                                              (self.N, self.N))])

        return corrmat

    def time_surrogate_for_mi(self, bins=16, sample_range=100, tau_max=1,
                              lag_mode='all'):
        """
        Returns a joint shuffled surrogate of the full dataarray of length
        sample_range for all taus.

        Used for time evolution analysis. First one initializes the
        CouplingAnalysis class with the full dataarray and then this function
        is called for every single surrogate.

        :arg int sample_range: length of sample
        :arg int bins: number of bins for estimating MI
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """

        if bins < 255:
            dtype = 'uint8'
        else:
            dtype = 'int16'

        perm = numpy.random.permutation(
            range(tau_max, self.total_time - tau_max))[:sample_range]

        # get the bin quantile steps
        bin_edge = numpy.ceil(sample_range/float(bins))

        symbolic_array = numpy.empty((2*tau_max + 1, self.N, sample_range),
                                     dtype=dtype)

        for t in range(2*tau_max + 1):
            tau = t - tau_max

            array = self.dataarray[:, perm + tau]

            # get the lower edges of the bins for every time series
            edges = numpy.sort(array, axis=1)[:, ::bin_edge]
            bins = edges.shape[1]

            # This gives the symbolic time series
            symbolic_array[t] = \
                (array.reshape(self.N, sample_range, 1)
                 >= edges.reshape(self.N, 1, bins)).sum(axis=2) - 1

        return self._calculate_mi(symbolic_array, corr_range=sample_range,
                                  bins=bins, tau_max=tau_max,
                                  lag_mode=lag_mode)

    def _calculate_mi(self, array, corr_range, bins, tau_max, lag_mode):
        """
        Returns the mi matrix.

        :arg int bins: number of bins for estimating MI
        :arg int tau_max: maximum lag in both directions, including last lag
        :arg str lag_mode: output mode
        :rtype: 3D numpy array (float) [index, index, index]
        :return: correlation matrix with different lag_mode choices
        """

        # lag_mode dict
        mode = self.lag_modi[lag_mode]
        only_tri = int(self.only_tri)

        # Initialize
        hist2D = numpy.zeros((bins, bins), dtype="int32")
        if lag_mode == 'all':
            corrmat = numpy.zeros((2*tau_max + 1, self.N, self.N),
                                  dtype='float32')
        elif lag_mode == 'sum':
            corrmat = numpy.zeros((2, self.N, self.N), dtype='float32')
        elif lag_mode == 'max':
            corrmat = numpy.zeros((2, self.N, self.N), dtype='float32')

        # Precalculation of the log
        gfunc = numpy.zeros(corr_range+1)
        for t in range(1, corr_range + 1):
            gfunc[t] = t*numpy.log(t)

        # loop over all node pairs, NOT symmetric due to time shifts!
        for i in range(self.N-only_tri):
            for j in range((i+1)*only_tri, self.N):

                if mode == 2:
                    maxcross = 0.0
                    argmax = 0

                # loop over taus from -tau_max to tau_max INCLUDING the last
                # tau value
                for t in range(2*tau_max + 1):
                    tau = t - tau_max

                    # here the joint probability distribution is calculated
                    for k in range(corr_range):
                        indexi = array[tau_max, i, k]
                        indexj = array[t, j, k]
                        hist2D[indexi, indexj] += 1

                    # here the joint entropy is calculated by summing over all
                    # pattern combinations
                    jointent = 0.0
                    for m in range(bins):
                        for n in range(bins):
                            jointent -= gfunc[hist2D[m, n]]
                            hist2D[m, n] = 0

                    jointent /= float(corr_range)
                    jointent += numpy.log(float(corr_range))

                    # Mutual Information is...
                    mi = 0.0
                    mi = 2. * numpy.log(bins) - jointent

                    # norm the mi
                    mi /= numpy.log(bins)

                    # fill in values in matrix depending on lag_mode
                    if mode == 0:
                        corrmat[t, i, j] = mi

                    elif mode == 1:
                        if t <= tau_max:
                            corrmat[1, i, j] += mi
                        if t >= tau_max:
                            corrmat[0, i, j] += mi

                    elif mode == 2:
                        # calculate max and argmax by comparing to previous
                        # value and storing max
                        if mi > maxcross:
                            maxcross = mi
                            argmax = tau

                if mode == 2:
                    corrmat[0, i, j] = maxcross
                    corrmat[1, i, j] = argmax

        if self.only_tri:
            if lag_mode == 'all':
                corrmat = corrmat + corrmat.transpose(0, 2, 1)[::-1]
            if lag_mode == 'sum':
                corrmat[0] += corrmat[1].transpose()
                corrmat[1] = corrmat[0].transpose()
            elif lag_mode == 'max':
                corrmat[0] += corrmat[0].transpose()
                corrmat[1] -= corrmat[1].transpose()

        return corrmat

    #
    # A subroutine for fourier surrogates (from J Donges)
    #

    def correlatedNoiseSurrogates(self, original):
        """
        Generates surrogates by Fourier transforming the original time series,
        randomizing the phases and then applying an inverse Fourier transform.
        Correlated noise surrogates share their power spectrum and
        autocorrelation function with the original time series.

        :type original: 2D array
        :arg original: dim. 0 is index of time series, dim. 1 is time
        :return: surrogate time series (same dimensions as original)
        """

        #  Calculate FFT of original time series
        #  The FFT of the original data has to be calculated only once, so it
        #  is stored in self.originalFFT
        if self.has_fft:
            surrogates = self.originalFFT
        else:
            surrogates = numpy.fft.fft(original, axis=1)
            self.originalFFT = surrogates
            self.has_fft = True

        (nNodes, ntime) = original.shape

        if (ntime % 2) == 0:
            lenPhase = (ntime - 2) / 2
        else:
            lenPhase = (ntime - 1) / 2

        #  Generate random phases uniformly distributed in the interval
        #  [0, 2*Pi]. Guarantee that the phases for positive and negative
        #  frquencies are the same to obtain real surrogates in the end!
        phases = numpy.random.uniform(low=0, high=2 * numpy.pi,
                                      size=(nNodes, lenPhase))

        #  Add random phases uniformly distributed in the interval [0, 2*Pi]
        surrogates[:, 1:lenPhase+1] *= numpy.exp(1j * phases)

        #  Discriminate between even and uneven number of samples
        #  Note that the output of fft has the following form:
        #  - Even sample number: (mean, pos. freq, nyquist freq, neg. freq)
        #  - Odd sample number: (mean, pos. freq, neg. freq)
        if (ntime % 2) == 0:
            surrogates[:, lenPhase+2:ntime] = \
                numpy.flipud(surrogates[:, 1:lenPhase+1].conjugate())
        else:
            surrogates[:, lenPhase+1:ntime] = \
                numpy.flipud(surrogates[:, 1:lenPhase+1].conjugate())

        #  Calculate IFFT and take the real part, the remaining imaginary part
        #  is due to numerical errors
        return numpy.ascontiguousarray(numpy.real(
            numpy.fft.ifft(surrogates, axis=1)))



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np

from ..core._ext.types import to_cy, FIELD
from ._ext.numerics import mutual_information
from .climate_data import ClimateData
from .climate_network import ClimateNetwork


class MutualInfoClimateNetwork(ClimateNetwork):
    """
    Represents a mutual information climate network.

    Constructs a static climate network based on mutual information at zero
    lag, as in [Ueoka2008]_.

    Mutual information climate networks are undirected, since mutual
    information is a symmetrical measure. In contrast to Pearson correlation
    used in :class:`.TsonisClimateNetwork`, mutual information has the
    potential to detect nonlinear statistical interdependencies.
    """

    #
    #  Defines internal methods
    #

    def __init__(self, data, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface", winter_only=True,
                 silence_level=0):
        """
        Initialize an instance of MutualInfoClimateNework.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :class:`.ClimateData`
        :arg data: The climate data used for network construction.

        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str node_weight_type: The type of geographical node weight to be
            used.
        :arg bool winter_only: Determines, whether only data points from the
            winter months (December, January and February) should be used for
            analysis. Possibly, this further suppresses the annual cycle in the
            time series.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a mutual information climate network...")
        self.silence_level = silence_level

        #  Set instance variables
        assert isinstance(data, ClimateData)
        self.data: ClimateData = data
        """The climate data used for network construction."""
        self.N = self.data.grid.N
        self._prescribed_link_density = link_density
        self._winter_only = winter_only

        #  Class specific settings
        self.mi_file = "mutual_information_" + data.data_source + "_" \
            + data.observable_name + ".data"
        """(string) - The name of the file for storing the mutual information
                      matrix."""

        self._set_winter_only(winter_only)
        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=self._similarity_measure,
                                threshold=threshold,
                                non_local=non_local,
                                directed=False,
                                node_weight_type=node_weight_type,
                                silence_level=silence_level)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        return ClimateNetwork.__cache_state__(self)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return ClimateNetwork.__rec_cache_state__(self) + (self.data,)

    def __str__(self):
        """
        Return a string representation of MutualInfoClimateNetwork.
        """
        return 'MutualInfoClimateNetwork:\n' + ClimateNetwork.__str__(self)

    def _cython_calculate_mutual_information(self, anomaly, n_bins=32):
        """
        Calculate the mutual information matrix at zero lag.

        The cython code is adopted from the Tisean 3.0.1 mutual.c module.

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: The anomaly time series.

        :arg int n_bins: The number of bins for estimating probability
            distributions.
        :arg bool fast: Indicates, whether fast or slow algorithm should be
            used.
        :rtype: 2D array (index, index)
        :return: the mutual information matrix at zero lag.
        """
        if self.silence_level <= 1:
            print("Calculating mutual information matrix at zero lag from "
                  "anomaly values using cython...")

        #  Normalize anomaly time series to zero mean and unit variance
        self.data.normalize_time_series_array(anomaly)

        #  Create local transposed copy of anomaly
        anomaly = anomaly.T.copy()

        (N, n_samples) = anomaly.shape

        #  Get common range for all histograms
        range_min = float(anomaly.min())
        range_max = float(anomaly.max())

        #  Rescale all time series to the interval [0,1],
        #  using the maximum range of the whole dataset.
        scaling = 1./(range_max - range_min)

        mi = mutual_information(
            to_cy(anomaly, FIELD), n_samples, N, n_bins, scaling, range_min)

        if self.silence_level <= 1:
            print("Done!")

        return mi

    def calculate_similarity_measure(self, anomaly):
        """
        Calculate the mutual information matrix.

        Encapsulates calculation of mutual information with standard
        parameters.

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: The anomaly time series.

        :rtype: 2D Numpy array (index, index)
        :return: the mutual information matrix at zero lag.
        """
        return self._cython_calculate_mutual_information(anomaly)

    def mutual_information(self, anomaly=None, dump=True):
        """
        Return mutual information matrix at zero lag.

        Check if mutual information matrix (MI) was already calculated before:
          - If yes, return MI from a data file.
          - If not, return MI from calculation and store in file.

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: The anomaly time series.
        :arg bool dump: Store MI in data file.

        :rtype: 2D Numpy array (index, index)
        :return: the mutual information matrix at zero lag.
        """
        try:
            #  Try to load MI from file
            if self.silence_level <= 1:
                print("Loading mutual information matrix from "
                      f"{self.mi_file}...")

            with open(self.mi_file, 'r', encoding="utf-8") as f:
                mi = np.load(f)
                #  Check if the dimensions of mutual_information correspond to
                #  the grid.
                if mi.shape != (self.N, self.N):
                    print(f"{self.mi_file} in current directory has "
                          "incorrect dimensions!")
                    raise RuntimeError

        except (IOError, RuntimeError):
            if self.silence_level <= 1:
                print("An error occured while loading data from "
                      f"{self.mi_file}.")
                print("Recalculating mutual information.")

            mi = self._cython_calculate_mutual_information(anomaly)
            if dump:
                with open(self.mi_file, 'w', encoding="utf-8") as f:
                    if self.silence_level <= 1:
                        print("Storing in", self.mi_file)
                    mi.dump(f)

        return mi

    def winter_only(self):
        """
        Indicate, if only winter months were used for network generation.

        :return bool: whether only winter months were used for network
            generation.
        """
        return self._winter_only

    def _set_winter_only(self, winter_only, dump=False):
        """
        Toggle use of exclusively winter data points for network generation.

        :arg bool winter_only: Indicates whether only winter months were used
            for network generation.
        :arg bool dump: Store MI in data file.
        """
        self._winter_only = winter_only
        if winter_only:
            winter_anomaly = self.data.anomaly_selected_months([0, 1, 11])
            mi = self.mutual_information(winter_anomaly, dump=dump)
        else:
            mi = self.mutual_information(self.data.anomaly(), dump=dump)
        self._similarity_measure = mi

    def set_winter_only(self, winter_only, dump=True):
        """
        Toggle use of exclusively winter data points for network generation.

        Also explicitly regenerates the instance of MutualInfoClimateNetwork.

        :arg bool winter_only: Indicates whether only winter months were used
            for network generation.
        :arg bool dump: Store MI in data file.
        """
        self._set_winter_only(winter_only, dump=dump)
        self._regenerate_network()

    #
    #  Defines methods to calculate  weighted network measures
    #

    def mutual_information_weighted_average_path_length(self):
        """
        Return mutual information weighted average path length.

        :return float: the mutual information weighted average path length.
        """
        return self._weighted_metric(
            "mutual_information",
            lambda: np.abs(self.mutual_information()),
            "average_path_length")

    def mutual_information_weighted_closeness(self):
        """
        Return mutual information weighted closeness.

        :rtype: 1D Numpy array [index]
        :return: the mutual information weighted closeness sequence.
        """
        return self._weighted_metric(
            "mutual_information",
            lambda: np.abs(self.mutual_information()),
            "closeness")

    def local_mutual_information_weighted_vulnerability(self):
        """
        Return mutual information weighted vulnerability.

        :rtype: 1D Numpy array [index]
        :return: the mutual information weighted vulnerability sequence.
        """
        return self._weighted_metric(
            "mutual_information",
            lambda: np.abs(self.mutual_information()),
            "local_vulnerability")



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for the analysis of dynamical systems and time series based
on event synchronization and event coincidence analysis
"""

# array object and fast numerics
import numpy as np

from ..eventseries import EventSeries
from .climate_network import ClimateNetwork
from .climate_data import ClimateData
from ..core import Data

#
#  Class definitions
#


class EventSeriesClimateNetwork(EventSeries, ClimateNetwork):
    """
    Class EventSeriesClimateNetwork for generating and quantitatively
    analyzing event synchronisation and event coincidence analysis networks

    References: [Boers2014]_.
    """

    #
    #  Internal methods

    def __init__(self, data, method='ES', p_value=None, **kwargs):

        r"""
        Initialize an instance of EventSeriesClimateNetwork.

        For other applications of event series networks please use
        the EventSeries class together with the Network class.

        :type data: :classL`..climate.ClimateData`
        :arg data: The climate data used for network construction.
        :type method: str {'ES', 'ECA', 'ES_pval', 'ECA_pval'}
        :arg method: determines if ES, ECA, or the p-values of one of the
                     methods should be used for network reconstruction.
        :type p_value: float in [0,1]
        :arg p_value: determines the p-value threshold for network
                      reconstruction. ES/ECA scores of event time series pairs
                      with p-value higher than threshold are set to zero
                      leading to missing link in climate network.
                      Default: None. No p-value thresholding.
        :arg \**kwargs:
            optional keyword arguments to specify parent classes' behavior,
            see below for all options.

        :Keyword Arguments:
            * *taumax* (``float``) --
              maximum time difference between two events to be
              considered synchronous. Caution: For ES, the default is
              np.inf because of the intrinsic dynamic coincidence
              interval in ES. For ECA, taumax is a parameter of the
              method that needs to be defined.
            * *lag* (``float``) --
              extra time lag between the event series.
            * *symmetrization* (``str {'directed', 'symmetric', 'antisym',
                                    'mean', 'max', 'min'}`` for ES,
                                ``str {'directed', 'mean', 'max', 'min'}``
                                for ECA) --
              determines if and if true, which symmetrization
              should be used for the ES/ECA score matrix.
            * *window_type* (``str {'retarded', 'advanced', 'symmetric'}``) --
              Only for ECA. Determines if precursor coincidence
              rate ('advanced'), trigger coincidence rate
              ('retarded') or a general coincidence rate with the
              symmetric interval [-taumax, taumax] are computed
              ('symmetric'). Default: 'symmetric'.
            * *threshold_method* (``str 'quantile' or 'value'
                                  or 1D numpy array of
                                  'quantile' or 'value'``) --
              specifies the method for generating a binary
              event matrix from an array of continuous time
              series. Default: None.
            * *threshold_values* (``1D Numpy array or float``) --
              quantile or real number determining threshold
              for each variable. Default: None.
            :type threshold_types: str 'above' or 'below'
                                or 1D list of strings
                                'above' or 'below'
            * *threshold_types* (``str 'above' or 'below'
                                 or 1D list of strings of
                                 'above' or 'below'``) --
              determines for each variable if event is below
              or above threshold.
            :type non_local: bool
            * *non_local* (``bool``) --
              determines whether links between spatially close
              nodes should be suppressed.
            :type node_weight_type: str
            * *node_weight_type* (``str``) --
              The type of geographical node weight to be
              used.
            * *arg silence_level* (``int``) --
              The inverse level of verbosity of the object.
        """

        # extract ES and CN related optional keyword arguments from **kwargs
        ES_kwargs = {
            "taumax": kwargs.get("taumax", np.inf),
            "lag": kwargs.get("lag", 0.0),
            "threshold_method": kwargs.get("threshold_method", None),
            "threshold_values": kwargs.get("threshold_values", None),
            "threshold_types": kwargs.get("threshold_types", None)
        }

        ES_analysis_kwargs = {
            "symmetrization": kwargs.get("symmetrization", 'directed'),
            "window_type": kwargs.get("window_type", 'symmetric')
        }

        ES_significance_kwargs = {
            "surrogate": kwargs.get("surrogate", 'shuffle'),
            "n_surr": kwargs.get("n_surr", 1000),
            "symmetrization": kwargs.get("symmetrization", 'directed'),
            "window_type": kwargs.get("window_type", 'symmetric')
        }

        CN_kwargs = {
            "non_local": kwargs.get("non_local", False),
            "node_weight_type": kwargs.get("node_weight_type", "surface"),
            "silence_level": kwargs.get("silence_level", 0)
        }

        method_types = ['ES', 'ECA', 'ES_pval', 'ECA_pval']
        if method not in method_types:
            raise IOError(f"Method input must be: "
                          f"{method_types[0]}, {method_types[1]},"
                          f"{method_types[2]}, or {method_types[3]}!")

        self.__method = method
        self.__p_value = p_value

        self.__symmetry = kwargs.get("symmetrization", 'directed')
        self.directed = self.__symmetry == "directed"

        # Construct an EventSeries object with the chosen parameters
        EventSeries.__init__(self, data.observable(), **ES_kwargs)

        # Compute matrix for link weights of ClimateNetwork from event
        # synchronization or event coincidence analysis with chosen symmetry
        # option
        measure_matrix = []

        # required here for bootstrapping the cache hierarchy
        self.grid = None

        # If standard ES/ECA measure is chosen, calculate pairwise ES/ECA
        # scores
        if self.__method in ['ES', 'ECA']:
            measure_matrix = \
                self.event_series_analysis(method=self.__method,
                                           **ES_analysis_kwargs)

            # Check if a p-value is chosen, then:
            # Set all coupling strengths in the measure matrix with
            # higher p-value to zero, leading to a missing edge in the climate
            # network
            if self.__p_value is not None:

                if self.__p_value > 1.0 or self.__p_value < 0.0:
                    raise IOError("'p_value' must lie in the unit interval!")

                significance_matrix = \
                    self.event_analysis_significance(
                        method=self.__method, **ES_significance_kwargs)

                for i in range(self.__N):
                    for j in range(self.__N):
                        if significance_matrix[i][j] < 1.0 - p_value:
                            measure_matrix[i][j] = 0.0

        elif self.__method in ['ES_pval', 'ECA_pval']:
            measure_matrix = \
                self.event_analysis_significance(
                    method=self.__method, **ES_significance_kwargs)

        ClimateNetwork.__init__(self, grid=data.grid,
                                similarity_measure=measure_matrix,
                                threshold=0, directed=self.directed,
                                **CN_kwargs)

    def __str__(self):
        """
        Return a string representation of EventSeriesClimateNetwork.

        **Example:**

        >>> data = EventSeriesClimateNetwork.SmallTestData()
        >>> print(EventSeriesClimateNetwork(data, taumax=16.0,
        >>>       threshold_method='quantile', threshold_value=0.8,
        >>>       threshold_types='above'))
        Extracting network adjacency matrix by thresholding...
        Setting area weights according to type surface ...
        Setting area weights according to type surface ...
        EventSeriesClimateNetwork:
        EventSeries: 6 variables, 10 timesteps, __taumax: 16.0, lag: 0.0
        ClimateNetwork:
        GeoNetwork:
        Network: directed, 6 nodes, 0 links, link density 0.000.
        Geographical boundaries:
                 time     lat     lon
           min    0.0    0.00    2.50
           max    9.0   25.00   15.00
        Threshold: 0
        Local connections filtered out: False
        Type of event series measure to construct
        the network: directedES
        """
        text = ("EventSeriesClimateNetwork:\n%s\n%s\n"
                "Type of event series measure to construct the network: "
                "%s%s")
        return text % (EventSeries.__str__(self),
                       ClimateNetwork.__str__(self), self.__symmetry,
                       self.__method)

    @staticmethod
    def SmallTestData():
        """
        Return test data set of 6 time series with 10 sampling points each.

        **Example:**

        >>> r(Data.SmallTestData().observable())
        array([[ 0.    ,  1.    ,  0.    , -1.    , -0.    ,  1.    ],
               [ 0.309 ,  0.9511, -0.309 , -0.9511,  0.309 ,  0.9511],
               [ 0.5878,  0.809 , -0.5878, -0.809 ,  0.5878,  0.809 ],
               [ 0.809 ,  0.5878, -0.809 , -0.5878,  0.809 ,  0.5878],
               [ 0.9511,  0.309 , -0.9511, -0.309 ,  0.9511,  0.309 ],
               [ 1.    ,  0.    , -1.    , -0.    ,  1.    ,  0.    ],
               [ 0.9511, -0.309 , -0.9511,  0.309 ,  0.9511, -0.309 ],
               [ 0.809 , -0.5878, -0.809 ,  0.5878,  0.809 , -0.5878],
               [ 0.5878, -0.809 , -0.5878,  0.809 ,  0.5878, -0.809 ],
               [ 0.309 , -0.9511, -0.309 ,  0.9511,  0.309 , -0.9511]])

        :rtype: ClimateData instance
        :return: a ClimateData instance for testing purposes.
        """
        data = Data.SmallTestData()

        return ClimateData(observable=data.observable(), grid=data.grid,
                           time_cycle=5, silence_level=2)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides classes for generating and analyzing complex climate networks.
"""

import numpy as np

from .climate_network import ClimateNetwork
from .climate_data import ClimateData
from ..core.cache import Cached


# TODO: Reconsider storage of correlation matrix without taking absolute value.
class TsonisClimateNetwork(ClimateNetwork):

    """
    Encapsulates a Tsonis climate network.

    Construct a static climate network following Tsonis et al. from the
    Pearson correlation matrix at zero lag.

    Hence, Tsonis climate networks are undirected due to the symmetry of the
    correlation matrix.

    References: [Tsonis2004]_, [Tsonis2006]_, [Tsonis2008b]_, [Tsonis2008c]_.
    """

    #
    #  Defines internal methods
    #

    def __init__(self, data, threshold=None, link_density=None,
                 non_local=False, node_weight_type="surface", winter_only=True,
                 silence_level=0):
        """
        Initialize an instance of TsonisClimateNetwork.

        .. note::
           Either threshold **OR** link_density have to be given!

        Possible choices for ``node_weight_type``:
          - None (constant unit weights)
          - "surface" (cos lat)
          - "irrigation" (cos**2 lat)

        :type data: :classL`.ClimateData`
        :arg data: The climate data used for network construction.
        :arg float threshold: The threshold of similarity measure, above which
            two nodes are linked in the network.
        :arg float link_density: The networks's desired link density.
        :arg bool non_local: Determines, whether links between spatially close
            nodes should be suppressed.
        :arg str ode_weight_type: The type of geographical node weight to be
            used.
        :arg bool winter_only: Determines, whether only data points from the
            winter months (December, January and February) should be used for
            analysis. Possibly, this further suppresses the annual cycle in the
            time series.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        if silence_level <= 1:
            print("Generating a Tsonis climate network...")
        self.silence_level = silence_level

        #  Set instance variables
        self.data = data
        """(ClimateData) - The climate data used for network construction."""
        self.N = self.data.grid.N

        self._set_winter_only(winter_only)
        ClimateNetwork.__init__(self, grid=self.data.grid,
                                similarity_measure=self.similarity_measure(),
                                threshold=threshold,
                                link_density=link_density,
                                non_local=non_local,
                                directed=False,
                                node_weight_type=node_weight_type,
                                silence_level=silence_level)

    def __str__(self):
        """
        Return a string representation of TsonisClimateNetwork.

        **Example:**

        >>> print(TsonisClimateNetwork.SmallTestNetwork())
        TsonisClimateNetwork:
        ClimateNetwork:
        GeoNetwork:
        Network: undirected, 6 nodes, 6 links, link density 0.400.
        Geographical boundaries:
                 time     lat     lon
           min    0.0    0.00    2.50
           max    9.0   25.00   15.00
        Threshold: 0.5
        Local connections filtered out: False
        Use only data points from winter months: False
        """
        return (f'TsonisClimateNetwork:\n{ClimateNetwork.__str__(self)}\n'
                f'Use only data points from winter months: '
                f'{self.winter_only()}')

    #
    #  Methods for testing purposes
    #

    @staticmethod
    def SmallTestNetwork():
        """
        Return a 6-node undirected test climate network from a test data set.

        **Example:**

        >>> r(TsonisClimateNetwork.SmallTestNetwork().adjacency)
        array([[0, 0, 1, 0, 1, 0], [0, 0, 0, 1, 0, 1], [1, 0, 0, 0, 1, 0],
               [0, 1, 0, 0, 0, 1], [1, 0, 1, 0, 0, 0], [0, 1, 0, 1, 0, 0]])

        :rtype: Network instance
        """
        return TsonisClimateNetwork(data=ClimateData.SmallTestData(),
                                    threshold=0.5, winter_only=False,
                                    silence_level=2)

    #
    #  Defines methods to calculate the correlation matrix
    #

    def _calculate_correlation(self, anomaly):
        """
        Return the correlation matrix at zero lag.

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: the anomaly time series from which to calculate the
                      correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        if self.silence_level <= 1:
            print("Calculating correlation matrix at zero lag from anomaly "
                  "values...")
        #  Cast to float32 type to save memory since correlation coefficients
        #  are not needed in high floating point precision.
        correlation = np.corrcoef(anomaly.transpose()).astype("float32")

        return correlation

    def calculate_similarity_measure(self, anomaly):
        """
        Encapsulate the calculation of the correlation matrix at zero lag.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork()._calculate_correlation(
        ...     anomaly=ClimateData.SmallTestData().anomaly())
        array([[ 1.        , -0.25377226, -1.        ,
                 0.25377226,  1.        , -0.25377226],
               [-0.25377226,  1.        ,  0.25377226,
                -1.        , -0.25377226,  1.        ],
               [-1.        ,  0.25377226,  1.        ,
                -0.25377226, -1.        ,  0.25377226],
               [ 0.25377226, -1.        , -0.25377226,
                 1.        ,  0.25377226, -1.        ],
               [ 1.        , -0.25377226, -1.        ,
                 0.25377226,  1.        , -0.25377226],
               [-0.25377226,  1.        ,  0.25377226,
                -1.        , -0.25377226,  1.        ]], dtype=float32)

        :type anomaly: 2D Numpy array (time, index)
        :arg anomaly: the anomaly time series from which to calculate the
                      correlation matrix at zero lag.

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        return self._calculate_correlation(anomaly)

    @Cached.method()
    def correlation(self):
        """
        Return the correlation matrix at zero lag.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork().correlation()
        array([[ 1.        ,  0.25377226,  1.        ,
                 0.25377226,  1.        ,  0.25377226],
               [ 0.25377226,  1.        ,  0.25377226,
                 1.        ,  0.25377226,  1.        ],
               [ 1.        ,  0.25377226,  1.        ,
                 0.25377226,  1.        ,  0.25377226],
               [ 0.25377226,  1.        ,  0.25377226,
                 1.        ,  0.25377226,  1.        ],
               [ 1.        ,  0.25377226,  1.        ,
                 0.25377226,  1.        ,  0.25377226],
               [ 0.25377226,  1.        ,  0.25377226,
                 1.        ,  0.25377226,  1.        ]], dtype=float32)

        :rtype: 2D Numpy array (index, index)
        :return: the correlation matrix at zero lag.
        """
        corr = self.similarity_measure()
        self.set_link_attribute('correlation', abs(corr))
        return corr

    def winter_only(self):
        """
        Indicate, if only winter months were used for network generation.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork().winter_only()
        False

        :return bool: whether only winter months were used for network
            generation.
        """
        return self._winter_only

    def _set_winter_only(self, winter_only):
        """
        Toggle use of exclusively winter data points for network generation.

        :arg bool winter_only: Indicates, whether only winter months were used
            for network generation.
        """
        self._winter_only = winter_only
        if winter_only:
            winter_anomaly = self.data.anomaly_selected_months([0, 1, 11])
            correlation = self._calculate_correlation(winter_anomaly)
        else:
            correlation = self._calculate_correlation(self.data.anomaly())
        self._similarity_measure = correlation

    def set_winter_only(self, winter_only):
        """
        Toggle use of exclusively winter data points for network generation.

        Also explicitly re(generates) the instance of TsonisClimateNetwork.

        **Example:**

        >>> net = TsonisClimateNetwork.SmallTestNetwork()
        >>> net.set_winter_only(winter_only=False)
        >>> net.n_links
        6

        :arg bool winter_only: Indicates, whether only winter months were used
            for network generation.
        """
        self._set_winter_only(winter_only)
        self._regenerate_network()

    #
    #  Defines methods to calculate  weighted network measures
    #

    def correlation_weighted_average_path_length(self):
        """
        Return correlation weighted average path length.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork().\
                correlation_weighted_average_path_length()
        1.0

        :return float: the correlation weighted average path length.
        """
        self.correlation()
        return self.average_path_length('correlation')

    def correlation_weighted_closeness(self):
        """
        Return correlation weighted closeness.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork().\
                correlation_weighted_closeness()
        array([ 0.25, 0.25, 0.25, 0.25, 0.25, 0.25])

        :rtype: 1D Numpy array [index]
        :return: the correlation weighted closeness sequence.
        """
        self.correlation()
        return self.closeness('correlation')

    def local_correlation_weighted_vulnerability(self):
        """
        Return correlation weighted vulnerability.

        **Example:**

        >>> TsonisClimateNetwork.SmallTestNetwork().\
                local_correlation_weighted_vulnerability()
        array([ 0., 0., 0., 0., 0., 0.])

        :rtype: 1D Numpy array [index]
        :return: the correlation weighted vulnerability sequence.
        """
        self.correlation()
        return self.local_vulnerability('correlation')



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Simple tests for the CoupledClimateNetwork class.
"""

from pyunicorn.climate import ClimateData, CoupledTsonisClimateNetwork


def test_internal_link_density():
    cd = ClimateData.SmallTestData()
    tsonis_ccn = CoupledTsonisClimateNetwork(cd, cd, threshold=.2)
    res = tsonis_ccn.internal_link_density()
    assert isinstance(res, tuple) and all(isinstance(d, float) for d in res)
    assert all(0 <= d <= 1 for d in res)



# Copyright (C) 2014 SWIPO Project
#
# Authors (this file):
#   Stefan Schinkel <stefan.schinkel@gmail.com>
"""
Provides Simple test for the treatment of complex admittances
in resistive networks.
"""

import numpy as np

from pyunicorn import ResNetwork


res = ResNetwork.SmallTestNetwork()
resC = ResNetwork.SmallComplexNetwork()
# a symmetric one too
r = np.zeros((5, 5), dtype=complex)
r.real = [[0, 2, 0, 0, 0],
          [2, 0, 8, 2, 0],
          [0, 8, 0, 8, 0],
          [0, 2, 8, 0, 10],
          [0, 0, 0, 10, 0]]
r.imag = r.real
resCsym = ResNetwork(r)


def testComplexAdmittance():
    # Ensure that the sum of real and imaginary
    # admittance values in resc is zeros

    sumAdm = resCsym.get_admittance().real + resCsym.get_admittance().imag
    assert np.all(sumAdm == 0)


def testPinvSymmetry():
    # test that the real and imag part of "R" are identical for
    # the test network

    assert np.allclose(resCsym.get_R().real, resCsym.get_R().imag)


def testAdmittiveDegree():

    sumAD = resCsym.admittive_degree().real + resCsym.admittive_degree().imag
    assert np.all(sumAD == 0)



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

from pathlib import Path

import matplotlib.pyplot as plt

from pyunicorn.climate.climate_data import ClimateData
from pyunicorn.climate.tsonis import TsonisClimateNetwork
from pyunicorn.climate.map_plot import MapPlot


# pylint: disable=too-few-public-methods
class TestMapPlot:
    """
    Simple tests for the `MapPlot` class.
    """

    @staticmethod
    def test_plot(reanalysis_data: Path):
        """
        Check error-free execution.
        """
        # prepare ClimateNetwork fixture
        # (select subset of data to speed up loading and calculation)
        title = "ncep_ncar_reanalysis"
        window = {
            "time_min": 0., "time_max": 0.,
            "lat_min": 30, "lon_min": 0,
            "lat_max": 50, "lon_max": 30}
        data = ClimateData.Load(
            file_name=str(reanalysis_data), observable_name="air",
            file_type="NetCDF", window=window, time_cycle=12)
        net = TsonisClimateNetwork(data, threshold=.05, winter_only=False)

        # create MapPlot
        map_plot = MapPlot(data.grid, title)
        assert map_plot.title == title

        # plot with suppressed display
        plt.ioff()
        map_plot.plot(net.degree(), "Degree")
        assert plt.gca().get_title() == title
        plt.close()



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Provides class for analyzing spatially embedded complex networks.
"""

from typing import Tuple
from collections.abc import Hashable

import numpy as np
from numpy import random
import igraph

from ._ext.types import to_cy, ADJ, NODE, FIELD, DEGREE
from ._ext.numerics import _randomly_rewire_geomodel_I, \
    _randomly_rewire_geomodel_II, _randomly_rewire_geomodel_III

from .network import Network
from .grid import Grid


class SpatialNetwork(Network):
    """
    Encapsulates a spatially embedded network.

    Adds more network measures and statistics based on the
    spatial embedding.
    """

    #
    #  Definitions of internal methods
    #

    def __init__(self, grid: Grid, adjacency=None, edge_list=None,
                 directed=False, silence_level=0):
        """
        Initialize an instance of SpatialNetwork.

        :type grid: :class:`.Grid`
        :arg grid: The Grid object describing the network's spatial embedding.
        :type adjacency: 2D array (int8) [index, index]
        :arg adjacency: The network's adjacency matrix.
        :type edge_list: array-like list of lists
        :arg  edge_list: Edge list of the new network.
                         Entries [i,0], [i,1] contain the end-nodes of an edge.
        :arg bool directed: Determines, whether the network is treated as
            directed.
        :arg int silence_level: The inverse level of verbosity of the object.
        """
        assert isinstance(grid, Grid)
        self.grid: Grid = grid
        """(Grid) - Grid object describing the network's spatial embedding"""

        #  Call constructor of parent class Network
        Network.__init__(self, adjacency=adjacency, edge_list=edge_list,
                         directed=directed, silence_level=silence_level)

    def __cache_state__(self) -> Tuple[Hashable, ...]:
        # The following attributes are assumed immutable:
        #   (grid)
        return Network.__cache_state__(self)

    def __rec_cache_state__(self) -> Tuple[object, ...]:
        return (self.grid,)

    def __str__(self):
        """
        Return a string representation of the SpatialNetwork object.
        """
        return f'SpatialNetwork:\n{Network.__str__(self)}'

    #
    #  Load and save GeoNetwork object
    #

    def save(self, filename, fileformat=None, *args, **kwds):
        """
        Save the SpatialNetwork object to files.

        Unified writing function for graphs. Relies on and partially extends
        the corresponding igraph function. Refer to igraph documentation for
        further details on the various writer methods for different formats.

        This method tries to identify the format of the graph given in
        the first parameter (based on extension) and calls the corresponding
        writer method.

        Existing node and link attributes/weights are also stored depending
        on the chosen file format. E.g., the formats GraphML and gzipped
        GraphML are able to store both node and link weights.

        The grid is not stored if the corresponding filename is None.

        The remaining arguments are passed to the writer method without
        any changes.

        :arg tuple/list filename: Tuple or list of two strings, namely
            the paths to the files where the Network object and the
            GeoGrid object are to be stored (filename_network, filename_grid)
        :arg str fileformat: the format of the file (if one wants to override
            the format determined from the filename extension, or the filename
            itself is a stream). ``None`` means auto-detection.  Possible
            values are: ``"ncol"`` (NCOL format), ``"lgl"`` (LGL format),
            ``"graphml"``, ``"graphmlz"`` (GraphML and gzipped GraphML format),
            ``"gml"`` (GML format), ``"dot"``, ``"graphviz"`` (DOT format, used
            by GraphViz), ``"net"``, ``"pajek"`` (Pajek format), ``"dimacs"``
            (DIMACS format), ``"edgelist"``, ``"edges"`` or ``"edge"`` (edge
            list), ``"adjacency"`` (adjacency matrix), ``"pickle"`` (Python
            pickled format), ``"svg"`` (Scalable Vector Graphics).
        """
        try:
            (filename_network, filename_grid) = filename
        except ValueError as e:
            raise ValueError("'filename' must be a tuple or list of two "
                             "items: filename_network, filename_grid") from e
        #  Store network
        Network.save(self, filename=filename_network, fileformat=fileformat,
                     *args, **kwds)

        #  Store grid
        if filename_grid is not None:
            self.grid.save(filename=filename_grid)

    @staticmethod
    def Load(filename, fileformat=None, silence_level=0, *args, **kwds):
        """
        Return a SpatialNetwork object stored in files.

        Unified reading function for graphs. Relies on and partially extends
        the corresponding igraph function. Refer to igraph documentation for
        further details on the various reader methods for different formats.

        This method tries to identify the format of the graph given in
        the first parameter and calls the corresponding reader method.

        Existing node and link attributes/weights are also restored depending
        on the chosen file format. E.g., the formats GraphML and gzipped
        GraphML are able to store both node and link weights.

        The remaining arguments are passed to the reader method without
        any changes.

        :arg tuple/list filename: Tuple or list of two strings, namely
            the paths to the files containing the Network object and the
            Grid object (filename_network, filename_grid)
        :arg str fileformat: the format of the file (if known in advance)
          ``None`` means auto-detection. Possible values are: ``"ncol"`` (NCOL
          format), ``"lgl"`` (LGL format), ``"graphml"``, ``"graphmlz"``
          (GraphML and gzipped GraphML format), ``"gml"`` (GML format),
          ``"net"``, ``"pajek"`` (Pajek format), ``"dimacs"`` (DIMACS format),
          ``"edgelist"``, ``"edges"`` or ``"edge"`` (edge list),
          ``"adjacency"`` (adjacency matrix), ``"pickle"`` (Python pickled
          format).
        :arg int silence_level: The inverse level of verbosity of the object.
        :rtype: SpatialNetwork object
        :return: :class:`SpatialNetwork` instance.
        """
        try:
            (filename_network, filename_grid) = filename
        except ValueError as e:
            raise ValueError("'filename' must be a tuple or list of two "
                             "items: filename_network, filename_grid") from e

        #  Load Grid object
        grid = Grid.Load(filename_grid)

        #  Load to igraph Graph object
        graph = igraph.Graph.Read(f=filename_network, format=fileformat,
                                  *args, **kwds)

        #  Extract adjacency matrix
        A = np.array(graph.get_adjacency(type=2).data)

        #  Create GeoNetwork instance
        net = SpatialNetwork(grid=grid, adjacency=A,
                             directed=graph.is_directed(),
                             silence_level=silence_level)

        #  Extract node weights
        if "node_weight_nsi" in graph.vs.attribute_names():
            node_weights = \
                np.array(graph.vs.get_attribute_values("node_weight_nsi"))
            net.node_weights = node_weights

        #  Overwrite igraph Graph object in Network instance to restore link
        #  attributes/weights
        net.graph = graph
        #  invalidate cache
        net._mut_la += 1
        return net

    @staticmethod
    def SmallTestNetwork():
        """
        Return a 6-node undirected geographically embedded test network.

        The test network consists of the SmallTestNetwork of the Network class
        with node coordinates given by the SmallTestGrid of the Grid class.

        The network looks like this::

                3 - 1
                |   | \\
            5 - 0 - 4 - 2

        :rtype: SpatialNetwork object
        :return: :class:`SpatialNetwork` instance for testing purposes.
        """
        return SpatialNetwork(grid=Grid.SmallTestGrid(),
                              adjacency=Network.SmallTestNetwork().adjacency,
                              directed=False, silence_level=2)

    @staticmethod
    def Model(network_model, grid, **kwargs):
        """
        Return a new model graph generated with the specified network model
        and embedded on the specified spatial grid
        """
        A = getattr(SpatialNetwork, network_model)(**kwargs)
        return SpatialNetwork(grid, A)

    #
    #  Graph randomization methods
    #

    #  TODO: Experimental code!
    def randomly_rewire_geomodel_I(self, distance_matrix, iterations,
                                   inaccuracy):
        """
        Randomly rewire the current network in place using geographical
        model I.

        Geographical model I preserves the degree sequence (exactly) and the
        link distance distribution :math:`p(l)` (approximately).

        A higher ``inaccuracy`` in the conservation of :math:`p(l)` will lead
        to

          - less deterministic links in the network and, hence,
          - more degrees of freedom for the random graph and
          - a shorter runtime of the algorithm, since more pairs of nodes
            eligible for rewiring can be found.

        **Example** (The degree sequence should be the same after rewiring):

        >>> net = SpatialNetwork.SmallTestNetwork()
        >>> net.randomly_rewire_geomodel_I(
        ...     distance_matrix=net.grid.distance(),
        ...     iterations=100, inaccuracy=100)
        #
        >>> net.degree()
        array([3, 3, 2, 2, 3, 1])

        :type distance_matrix: 2D Numpy array [index, index]
        :arg distance_matrix: Suitable distance matrix between nodes.

        :type iterations: number (int)
        :arg iterations: The number of rewirings to be performed.

        :type inaccuracy: number (float)
        :arg inaccuracy: The inaccuracy with which to conserve :math:`p(l)`.
        """
        if self.silence_level <= 1:
            print("Randomly rewiring given graph, preserving the degree "
                  "sequence and link distance distribution...")
        #  Get number of links
        E = self.n_links
        #  Collect adjacency and distance matrices
        A = to_cy(self.adjacency, ADJ)
        D = to_cy(distance_matrix, FIELD)
        #  Get degree sequence
        # degree = self.degree()

        #  Define for brevity
        eps = float(inaccuracy)

        # iterations = int(iterations)

        #  Get edge list
        edges = to_cy(np.array(self.graph.get_edgelist()), NODE)

        _randomly_rewire_geomodel_I(iterations, eps, A, D, E, edges)

        #  Update all other properties of GeoNetwork
        self.adjacency = A

    #  TODO: Experimental code!
    def randomly_rewire_geomodel_II(self, distance_matrix,
                                    iterations, inaccuracy):
        """
        Randomly rewire the current network in place using geographical
        model II.

        Geographical model II preserves the degree sequence :math:`k_v`
        (exactly), the link distance distribution :math:`p(l)` (approximately),
        and the average link distance sequence :math:`<l>_v` (approximately).

        A higher ``inaccuracy`` in the conservation of :math:`p(l)` and
        :math:`<l>_v` will lead to:

          - less deterministic links in the network and, hence,
          - more degrees of freedom for the random graph and
          - a shorter runtime of the algorithm, since more pairs of nodes
            eligible for rewiring can be found.

        :type distance_matrix: 2D Numpy array [index, index]
        :arg distance_matrix: Suitable distance matrix between nodes.

        :type iterations: number (int)
        :arg iterations: The number of rewirings to be performed.

        :type inaccuracy: number (float)
        :arg inaccuracy: The inaccuracy with which to conserve :math:`p(l)`.
        """
        #  FIXME: Add example
        if self.silence_level <= 1:
            print("Randomly rewiring given graph, preserving the degree "
                  "sequence, link distance distribution and average link "
                  "distance sequence...")

        #  Get number of links
        E = int(self.n_links)
        #  Collect adjacency and distance matrices
        A = to_cy(self.adjacency, ADJ)
        D = to_cy(distance_matrix, FIELD)

        #  Define for brevity
        eps = float(inaccuracy)

        #  Get edge list
        edges = to_cy(np.array(self.graph.get_edgelist()), NODE)

        _randomly_rewire_geomodel_II(iterations, eps, A, D, E, edges)

        #  Set new adjacency matrix
        self.adjacency = A

    #  TODO: Experimental code!
    def randomly_rewire_geomodel_III(self, distance_matrix,
                                     iterations, inaccuracy):
        """
        Randomly rewire the current network in place using geographical
        model III.

        Geographical model III preserves the degree sequence :math:`k_v`
        (exactly), the link distance distribution :math:`p(l)` (approximately),
        and the average link distance sequence :math:`<l>_v` (approximately).
        Moreover, degree-degree correlations are also conserved exactly.

        A higher ``inaccuracy`` in the conservation of :math:`p(l)` and
        :math:`<l>_v` will lead to:

          - less deterministic links in the network and, hence,
          - more degrees of freedom for the random graph and
          - a shorter runtime of the algorithm, since more pairs of nodes
            eligible for rewiring can be found.

        :type distance_matrix: 2D Numpy array [index, index]
        :arg distance_matrix: Suitable distance matrix between nodes.

        :type iterations: number (int)
        :arg iterations: The number of rewirings to be performed.

        :type inaccuracy: number (float)
        :arg inaccuracy: The inaccuracy with which to conserve :math:`p(l)`.
        """
        #  FIXME: Add example
        if self.silence_level <= 1:
            print("Randomly rewiring given graph, preserving the degree "
                  "sequence, degree-degree correlations, link distance "
                  "distribution and average link distance sequence...")

        #  Get number of nodes
        #  Get number of links
        E = int(self.n_links)
        #  Collect adjacency and distance matrices
        A = to_cy(self.adjacency, ADJ)
        D = to_cy(distance_matrix, FIELD)
        #  Get degree sequence
        degree = to_cy(self.degree(), DEGREE)

        #  Define for brevity
        eps = float(inaccuracy)

        #  Get edge list
        edges = to_cy(np.array(self.graph.get_edgelist()), NODE)

        _randomly_rewire_geomodel_III(iterations, eps, A, D, E, edges, degree)

        #  Set new adjacency matrix
        self.adjacency = A

    def set_random_links_by_distance(self, a, b):
        """
        Reassign links independently with
        link probability = :math:`exp(a + b*angular distance)`.

        .. note::
           Modifies network in place, creates an undirected network!

        **Example** (Repeat until a network with 5 links is created):

        >>> net = GeoNetwork.SmallTestNetwork()
        >>> while (net.n_links != 5):
        ...     net.set_random_links_by_distance(a=0., b=-4.)
        >>> net.n_links
        5

        :type a: number (float)
        :arg a: The a parameter.

        :type b: number (float)
        :arg b: The b parameter.
        """
        #  Get angular distance matrix
        D = self.grid.distance()
        #  Calculate link probabilities
        p = np.exp(a + b * D)

        #  Generate random numbers
        P = random.random(D.shape)
        #  Symmetrize
        P = 0.5 * (P + P.transpose())

        #  Create new adjacency matrix
        A = (p >= P).astype(ADJ)
        #  Set diagonal to zero - no self-loops!
        np.fill_diagonal(A, 0)

        #  Set new adjacency matrix
        self.adjacency = A

    #
    #  Get link distance distribution
    #

    def link_distance_distribution(self, n_bins, grid_type="euclidean",
                                   geometry_corrected=False):
        """
        Return the normalized link distance distribution.

        Correct for the geometry of the embedding space by default.

        **Examples:**

        >>> GeoNetwork.SmallTestNetwork().link_distance_distribution(
        ...     n_bins=4, geometry_corrected=False)[0]
        array([ 0.14285714,  0.28571429,  0.28571429,  0.28571429])
        >>> GeoNetwork.SmallTestNetwork().link_distance_distribution(
        ...     n_bins=4, geometry_corrected=True)[0]
        array([ 0.09836066,  0.24590164,  0.32786885,  0.32786885])

        :arg int n_bins: The number of bins for histogram.
        :arg str grid_type: Type of grid, used for distance calculation, can
            take values "euclidean" and "spherical" (only for GeoNetwork).
        :arg bool geometry_corrected: Toggles correction for grid geometry.
        :rtype: tuple of three 1D arrays [bin]
        :return: the link distance distribution, statistical error,
                 and lower bin boundaries.
        """
        if self.silence_level <= 1:
            print("Calculating link distance distribution...")

        #  Collect matrices
        A = self.adjacency
        if grid_type == "spherical":
            if self.grid.__class__.__name__ != "GeoGrid":
                raise NotImplementedError("Spherical coordinates are only "
                                          "supported for GeoGrid!")
            D = self.grid.angular_distance()
        elif grid_type == "euclidean":
            D = self.grid.euclidean_distance()
        else:
            raise ValueError("Grid type unknown!")

        #  Determine range for link distance histograms
        interval = (0, D.max())

        #  Get link distance distribution
        (dist, error, lbb) = self._histogram(D[A == 1], n_bins=n_bins,
                                             interval=interval)

        if geometry_corrected:
            geometric_ld_dist = \
                self.grid.geometric_distance_distribution(n_bins)[0]
            # Divide out the geometrical factor of the distribution
            dist /= geometric_ld_dist

        #  Normalize the distribution
        dist /= dist.sum()

        return (dist, error, lbb)

    #
    #  Distance related measures
    #

    #  (Un)directed average link distances

    def _calculate_general_average_link_distance(self, adjacency, degrees,
                                                 geometry_corrected=False):
        """
        Return general average link distances (:math:`ALD`).

        This general method is called to calculate undirected average link
        distance, average in-link distance and average out-link distance.

        The resulting sequence can optionally be corrected for biases in
        average link distance arising due to the grid geometry. E.g., for
        regional networks, nodes on the boundaries may have a bias towards
        larger values of :math:`ALD`, while nodes in the center have a bias
        towards smaller values of :math:`ALD`.

        :type adjacency: 2D array [index, index]
        :arg adjacency: The adjacency matrix.
        :type degrees: 1D array [index]
        :arg degrees: The degree sequence.
        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the general average link distance sequence.
        """
        D = self.grid.distance()
        k = self.degree()

        average_link_distance = np.zeros(self.N)

        #  Normalize by degree, not by number of nodes!!!
        average_link_distance[k != 0] = \
            (D * adjacency).sum(axis=1)[k != 0] / k[k != 0]

        if geometry_corrected:
            #  Calculate the average link distance for a fully connected
            #  network to correct for geometrical biases, particularly in
            #  regional networks.
            ald_correction = D.mean(axis=1)

            #  Correct average link distance
            average_link_distance /= ald_correction

        return average_link_distance

    def average_link_distance(self, geometry_corrected=False):
        """
        Return average link distances (undirected).

        .. note::
           Does not use directionality information.

        **Examples:**

        >>> SpatialNetwork.SmallTestNetwork().\
                average_link_distance(geometry_corrected=False)
        array([22.36067963, 11.18033981,  8.38525486, 13.97542477, 16.77050908,
               27.95084953])
        >>> SpatialNetwork.SmallTestNetwork().\
                average_link_distance(geometry_corrected=True)[:-1]
        array([1.6       , 1.09090909, 1.        , 1.66666667, 1.63636357])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the average link distance sequence (undirected).
        """
        if self.silence_level <= 1:
            print("Calculating average link distance...")

        A = self.undirected_adjacency().A
        degree = self.degree()

        return self._calculate_general_average_link_distance(
            A, degree, geometry_corrected=geometry_corrected)

    def inaverage_link_distance(self, geometry_corrected=False):
        """
        Return in-average link distances.

        Return regular average link distance for undirected networks.

        **Example:**

        >>> SpatialNetwork.SmallTestNetwork().\
                inaverage_link_distance(geometry_corrected=False)
        array([22.36067963, 11.18033981,  8.38525486, 13.97542477, 16.77050908,
               27.95084953])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the in-average link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating in-average link distance...")

        A = self.adjacency.T
        in_degree = self.indegree()

        return self._calculate_general_average_link_distance(
            A, in_degree, geometry_corrected=geometry_corrected)

    def outaverage_link_distance(self, geometry_corrected=False):
        """
        Return out-average link distances.

        Return regular average link distance for undirected networks.

        **Example:**

        >>> SpatialNetwork.SmallTestNetwork().
                outaverage_link_distance(geometry_corrected=False)
        array([22.36067963, 11.18033981,  8.38525486, 13.97542477, 16.77050908,
               27.95084953])

        :arg bool geometry_corrected: Toggles geometry correction.
        :rtype: 1D array [index]
        :return: the out-average link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating out-average link distance...")

        A = self.adjacency
        out_degree = self.outdegree()

        return self._calculate_general_average_link_distance(
            A, out_degree, geometry_corrected=geometry_corrected)

    def max_link_distance(self):
        """
        Return maximum angular geodesic link distances.

        .. note::
           Does not use directionality information.

        **Example:**

        >>> SpatialNetwork.SmallTestNetwork().max_link_distance()
        array([27.95085, 16.77051, 11.18034, 16.77051, 22.36068, 27.95085],
              dtype=float32)

        :rtype: 1D Numpy array [index]
        :return: the maximum link distance sequence.
        """
        if self.silence_level <= 1:
            print("Calculating maximum link distance...")

        A = self.undirected_adjacency().A
        D = self.grid.distance()

        maximum_link_distance = (D * A).max(axis=1)
        return maximum_link_distance

    #
    #  Link weighted network measures
    #

    def distance(self):
        """
        Return the distance matrix.
        """
        dist = self.grid.distance()
        if not self.find_link_attribute('distance'):
            self.set_link_attribute('distance', dist)
        return dist

    def average_distance_weighted_path_length(self):
        """
        Return average distance weighted path length.

        Returns the average path length link-weighted by the angular
        great circle distance between nodes.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                average_distance_weighted_path_length())
        0.4985

        :rtype: number (float)
        :return: the average distance weighted path length.
        """
        self.distance()
        return self.average_path_length('distance')

    def distance_weighted_closeness(self):
        """
        Return distance weighted closeness.

        Returns the sequence of closeness centralities link-weighted by the
        angular great circle distance between nodes.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                distance_weighted_closeness())
        array([ 2.2378, 2.4501, 2.2396, 2.4501, 2.2396, 1.1982])

        :rtype: 1D Numpy array [index]
        :return: the distance weighted closeness sequence.
        """
        self.distance()
        return self.closeness('distance')

    def local_distance_weighted_vulnerability(self):
        """
        Return local distance weighted vulnerability.

        Return the sequence of vulnerabilities link-weighted by
        the angular great circle distance between nodes.

        **Example:**

        >>> r(GeoNetwork.SmallTestNetwork().\
                local_distance_weighted_vulnerability())
        array([ 0.0325, 0.3137, 0.2056, 0.028 , -0.0283, -0.288 ])

        :rtype: 1D Numpy array [index]
        :return: the local distance weighted vulnerability sequence.
        """
        self.distance()
        return self.local_vulnerability('distance')



# This file is part of pyunicorn.
# Copyright (C) 2008--2024 Jonathan F. Donges and pyunicorn authors
# URL: <https://www.pik-potsdam.de/members/donges/software-2/software>
# License: BSD (3-clause)
#
# Please acknowledge and cite the use of this software and its authors
# when results are used in publications or published elsewhere.
#
# You can use the following reference:
# J.F. Donges, J. Heitzig, B. Beronov, M. Wiedermann, J. Runge, Q.-Y. Feng,
# L. Tupikina, V. Stolbova, R.V. Donner, N. Marwan, H.A. Dijkstra,
# and J. Kurths, "Unified functional network and nonlinear time series analysis
# for complex systems science: The pyunicorn package"

"""
Tests for the EventSeriesClimateNetwork class.
"""
import numpy as np

from pyunicorn.core.data import Data
from pyunicorn.climate.eventseries_climatenetwork import\
 EventSeriesClimateNetwork


def test_str(capsys):
    data = EventSeriesClimateNetwork.SmallTestData()
    print(EventSeriesClimateNetwork(data, method='ES',
                                    threshold_method='quantile',
                                    threshold_values=0.8, taumax=16,
                                    threshold_types='above'))
    out, err = capsys.readouterr()
    out_ref = "Extracting network adjacency matrix by thresholding...\n" + \
              "Setting area weights according to type surface ...\n" + \
              "Setting area weights according to type surface ...\n" + \
              "EventSeriesClimateNetwork:\n" + \
              "EventSeries: 6 variables, 10 timesteps, taumax: 16.0, " \
              "lag: 0.0" + \
              "\nClimateNetwork:\n" + \
              "GeoNetwork:\n" + \
              "SpatialNetwork:\n" + \
              "Network: directed, 6 nodes, 0 links, link density 0.000.\n" + \
              "Geographical boundaries:\n" + \
              "         time     lat     lon\n" + \
              "   min    0.0    0.00    2.50\n" + \
              "   max    9.0   25.00   15.00\n" + \
              "Threshold: 0\n" + \
              "Local connections filtered out: False\n" + \
              "Type of event series measure to construct " + \
              "the network: directedES\n"
    assert out == out_ref


def test_SmallTestData():
    res = Data.SmallTestData().observable()
    exp = np.array([[0., 1., 0., -1., -0., 1.],
                    [0.309, 0.9511, -0.309, -0.9511, 0.309, 0.9511],
                    [0.5878, 0.809, -0.5878, -0.809, 0.5878, 0.809],
                    [0.809, 0.5878, -0.809, -0.5878, 0.809, 0.5878],
                    [0.9511, 0.309, -0.9511, -0.309, 0.9511, 0.309],
                    [1., 0., -1., -0., 1., 0.],
                    [0.9511, -0.309, -0.9511, 0.309, 0.9511, -0.309],
                    [0.809, -0.5878, -0.809, 0.5878, 0.809, -0.5878],
                    [0.5878, -0.809, -0.5878, 0.809, 0.5878, -0.809],
                    [0.309, -0.9511, -0.309, 0.9511, 0.309, -0.9511]])
    assert np.allclose(res, exp, atol=1e-04)
